# Generated by Django 3.2.24 on 2025-06-28 14:07

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import educommon.django.db.mixins
import educommon.utils.date


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('function_tools', '0005_auto_20220724_0050'),
        ('edu_rdm_integration', '0017_delete_uploaddatacommand'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='CollectingDataStageStatus',
                    fields=[
                        ('title', models.TextField(verbose_name='расшифровка значения')),
                        (
                            'key',
                            models.CharField(
                                db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ'
                            ),
                        ),
                        ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
                    ],
                    options={
                        'verbose_name': 'Модель-перечисление статусов этапа сбора данных',
                        'verbose_name_plural': 'Модели-перечисления статусов этапов сбора данных',
                        'db_table': 'rdm_collecting_data_stage_status',
                    },
                ),
                migrations.CreateModel(
                    name='CollectingDataSubStageStatus',
                    fields=[
                        ('title', models.TextField(verbose_name='расшифровка значения')),
                        (
                            'key',
                            models.CharField(
                                db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ'
                            ),
                        ),
                        ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
                    ],
                    options={
                        'verbose_name': 'Модель-перечисление статусов подэтапа сбора данных',
                        'verbose_name_plural': 'Модели-перечисления статусов подэтапов сбора данных',
                        'db_table': 'rdm_collecting_data_sub_stage_status',
                    },
                ),
                migrations.CreateModel(
                    name='CollectingExportedDataStage',
                    fields=[
                        (
                            'id',
                            models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
                        ),
                        (
                            'logs_period_started_at',
                            models.DateTimeField(
                                db_index=True,
                                default=educommon.utils.date.get_today_min_datetime,
                                verbose_name='Левая граница периода обрабатываемых логов',
                            ),
                        ),
                        (
                            'logs_period_ended_at',
                            models.DateTimeField(
                                db_index=True,
                                default=educommon.utils.date.get_today_max_datetime,
                                verbose_name='Правая граница периода обрабатываемых логов',
                            ),
                        ),
                        (
                            'started_at',
                            models.DateTimeField(
                                auto_now_add=True, db_index=True, verbose_name='Время начала сбора данных'
                            ),
                        ),
                        (
                            'ended_at',
                            models.DateTimeField(
                                blank=True, db_index=True, null=True, verbose_name='Время завершения сбора данных'
                            ),
                        ),
                        (
                            'manager',
                            models.ForeignKey(
                                blank=True,
                                null=True,
                                on_delete=django.db.models.deletion.PROTECT,
                                to='function_tools.entity',
                                verbose_name='Менеджер ранера Функции',
                            ),
                        ),
                        (
                            'status',
                            models.ForeignKey(
                                default='CREATED',
                                on_delete=django.db.models.deletion.PROTECT,
                                to='edu_rdm_integration_collect_data_stage.collectingdatastagestatus',
                                verbose_name='Статус',
                            ),
                        ),
                    ],
                    options={
                        'verbose_name': 'Этап формирования данных для выгрузки',
                        'verbose_name_plural': 'Этапы формирования данных для выгрузки',
                        'db_table': 'rdm_collecting_exported_data_stage',
                    },
                    bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
                ),
                migrations.CreateModel(
                    name='CollectingExportedDataSubStage',
                    fields=[
                        (
                            'id',
                            models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
                        ),
                        (
                            'started_at',
                            models.DateTimeField(
                                auto_now_add=True, db_index=True, verbose_name='Время начала сбора данных'
                            ),
                        ),
                        (
                            'ended_at',
                            models.DateTimeField(
                                blank=True, db_index=True, null=True, verbose_name='Время завершения сбора данных'
                            ),
                        ),
                        (
                            'function',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.PROTECT,
                                to='function_tools.entity',
                                verbose_name='Функция',
                            ),
                        ),
                        (
                            'previous',
                            models.ForeignKey(
                                blank=True,
                                null=True,
                                on_delete=django.db.models.deletion.CASCADE,
                                to='edu_rdm_integration_collect_data_stage.collectingexporteddatasubstage',
                                verbose_name='Предыдущий сбор данных',
                            ),
                        ),
                        (
                            'stage',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.PROTECT,
                                to='edu_rdm_integration_collect_data_stage.collectingexporteddatastage',
                                verbose_name='Этап подготовки данных для экспорта',
                            ),
                        ),
                        (
                            'status',
                            models.ForeignKey(
                                default='CREATED',
                                on_delete=django.db.models.deletion.PROTECT,
                                to='edu_rdm_integration_collect_data_stage.collectingdatasubstagestatus',
                                verbose_name='Статус',
                            ),
                        ),
                    ],
                    options={
                        'verbose_name': 'Подэтап формирования данных для выгрузки',
                        'verbose_name_plural': 'Подэтапы формирования данных для выгрузки',
                        'db_table': 'rdm_collecting_exported_data_sub_stage',
                    },
                    bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
                ),
            ],
            database_operations=[],
        ),
    ]



# Generated by Django 3.2.24 on 2025-07-04 08:10

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration_collect_data_stage', '0003_auto_20250704_0810'),
        ('edu_rdm_integration_models', '0001_initial'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='RegionalDataMartModelEnum',
            new_name='RDMModelEnum',
        ),
    ]



from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
)

from django.conf import (
    settings,
)
from django.db.models import (
    CharField,
    OuterRef,
    Subquery,
    Value,
)
from django.db.models.functions import (
    Cast,
    Coalesce,
    Least,
)
from django.db.transaction import (
    atomic,
)

from educommon import (
    logger,
)

from edu_rdm_integration.stages.collect_data.models import (
    RDMCollectingDataStage,
    RDMCollectingDataStageStatus,
    RDMCollectingDataSubStage,
    RDMCollectingDataSubStageStatus,
)


if TYPE_CHECKING:
    from edu_rdm_integration.rdm_models.models import (
        BaseRDMModel,
    )
    from edu_rdm_integration.stages.collect_data.functions.non_calculated.base.managers import (
        BaseCollectingExportedDataRunnerManager,
    )


@atomic
def set_failed_status_suspended_collecting_data_stages() -> dict[str, int]:
    """Установить статус 'Завершено с ошибками' для зависших этапов и подэтапов сбора.

    Сборка считается зависшей в случае если за определенное в параметре RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT время,
    отсутствуют изменения в связанных подэтапах. Параметр RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT определяется
    в настройках приложения.
    """
    changed_status_result = {
        'change_stage_count': 0,
        'change_sub_stage_count': 0,
    }

    current_datetime = datetime.now()
    suspended_time_at = current_datetime - timedelta(minutes=settings.RDM_CHECK_SUSPEND_TASK_STAGE_TIMEOUT)

    suspended_stage_ids = set(
        RDMCollectingDataStage.objects.annotate(
            last_sub_stage_started_at=Coalesce(
                Subquery(
                    RDMCollectingDataSubStage.objects.filter(stage_id=OuterRef('pk'))
                    .values('started_at')
                    .order_by('-started_at')[:1]
                ),
                Value(datetime.combine(date.min, time.min)),
            )
        )
        .filter(
            last_sub_stage_started_at__lt=suspended_time_at,
            status__in=(
                RDMCollectingDataStageStatus.CREATED.key,
                RDMCollectingDataStageStatus.IN_PROGRESS.key,
            ),
        )
        .values_list('pk', flat=True)
    )

    if suspended_stage_ids:
        logger.info(f'find suspended CollectingDataStage: {", ".join(map(str, suspended_stage_ids))}..')

        change_stage_count = RDMCollectingDataStage.objects.filter(
            pk__in=suspended_stage_ids,
        ).update(
            status=RDMCollectingDataStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        change_sub_stage_count = RDMCollectingDataSubStage.objects.filter(
            stage_id__in=suspended_stage_ids,
        ).update(
            status=RDMCollectingDataSubStageStatus.FAILED.key,
            ended_at=current_datetime,
        )

        changed_status_result.update(
            {
                'change_stage_count': change_stage_count,
                'change_sub_stage_count': change_sub_stage_count,
            }
        )

    return changed_status_result


def get_collecting_managers_max_period_ended_dates(
    collecting_managers: Iterable['BaseCollectingExportedDataRunnerManager'],
) -> dict[str, 'datetime']:
    """Возвращает дату и время завершения последнего успешного этапа сбора для менеджеров Функций сбора."""
    managers_last_period_ended = (
        RDMCollectingDataStage.objects.filter(
            manager_id__in=[manager.uuid for manager in collecting_managers],
            id=Subquery(
                RDMCollectingDataStage.objects.filter(
                    manager_id=OuterRef('manager_id'),
                    status_id=RDMCollectingDataStageStatus.FINISHED.key,
                )
                .order_by('-id')
                .values('id')[:1]
            ),
        )
        .annotate(
            str_manager_id=Cast('manager_id', output_field=CharField()),
            last_period_ended_at=Least('logs_period_ended_at', 'started_at'),
        )
        .values_list(
            'str_manager_id',
            'last_period_ended_at',
        )
    )

    return {manager_id: last_period_ended_at for manager_id, last_period_ended_at in managers_last_period_ended}


def update_fields(entity: 'BaseRDMModel', field_values: dict[str, Any], mapping: dict[str, str]) -> None:
    """Обновление значений полей сущности по измененным полям модели.

    :param entity: Выгружаемая сущность
    :param field_values: Словарь с измененными данными модели
    :param mapping: Словарь маппинга полей модели к полям сущности
    """
    for model_field, entity_field in mapping.items():
        if model_field in field_values:
            setattr(entity, entity_field, field_values[model_field])



from edu_function_tools.validators import (
    EduFunctionValidator,
    EduRunnerValidator,
)


class BaseCollectingCalculatedExportedDataRunnerValidator(EduRunnerValidator):
    """Базовый класс валидаторов ранеров функций сбора расчетных данных для интеграции с РВД."""

    def validate(self, runnable):
        """Расширение метода валидации."""
        super().validate(runnable=runnable)


class BaseCollectingCalculatedExportedDataFunctionValidator(EduFunctionValidator):
    """Базовый класс валидаторов функций сбора расчетных данных для интеграции с РВД."""

    def validate(self, runnable):
        """Расширение метода валидации."""
        super().validate(runnable=runnable)



import asyncio
from abc import (
    ABCMeta,
)
from typing import (
    TYPE_CHECKING,
    Optional,
)

import asyncpg
from django.conf import (
    settings,
)
from django.db import (
    connection,
)
from django.db.models import (
    ForeignKey,
)

from educommon import (
    logger,
)
from m3_db_utils.models import (
    FictiveForeignKeyField,
)

from edu_rdm_integration.rdm_models.models import (
    BaseMainRDMModel,
    RDMModelEnum,
)


if TYPE_CHECKING:
    from asyncpg import (
        Pool,
    )

    from m3_db_utils.models import (
        ModelEnumValue,
    )


class BaseModelOutdatedDataCleaner(metaclass=ABCMeta):
    """Базовый класс уборщика устаревших данных моделей РВД."""

    # Запрос для разбиения таблицы на чанки и получения идентификаторов первых и последних записей
    SELECT_RDM_CHUNK_BOUNDED_SQL = """
            DO $$
            DECLARE
                chunk_size INT := {chunk_size};
                last_id INT := 0;
                first_id INT;
                last_chunk_id INT;
            BEGIN
                -- Создаем временную таблицу без ON COMMIT DROP
                DROP TABLE IF EXISTS rdm_chunk_bounds;
                CREATE TEMP TABLE rdm_chunk_bounds (
                    chunk_number INT,
                    first_id INT,
                    last_id INT
                );

                DROP TABLE IF EXISTS tmp_chunk;
                CREATE TEMP TABLE tmp_chunk (id INT) ON COMMIT DROP;

                WHILE TRUE LOOP
                    TRUNCATE tmp_chunk;

                    INSERT INTO tmp_chunk (id)
                    SELECT id
                    FROM {table_name}
                    WHERE id > last_id
                    ORDER BY id
                    LIMIT chunk_size;

                    IF NOT FOUND THEN
                        EXIT;
                    END IF;

                    SELECT MIN(id), MAX(id)
                    INTO first_id, last_chunk_id
                    FROM tmp_chunk;

                    INSERT INTO rdm_chunk_bounds (chunk_number, first_id, last_id)
                    VALUES (
                        (SELECT COUNT(*) FROM rdm_chunk_bounds) + 1,
                        first_id,
                        last_chunk_id
                    );

                    last_id := last_chunk_id;
                END LOOP;
            END $$;

            -- Теперь можно безопасно выбрать данные
            SELECT * FROM rdm_chunk_bounds ORDER BY chunk_number;
        """

    REMOVE_OUTDATED_DATA_SQL = """
        WITH deleted_rows AS (
            DELETE FROM {table_name}
            WHERE id IN (
                WITH tbl AS (
                    SELECT *
                    FROM {table_name}
                    WHERE id >= {first_id}
                        AND id <= {last_id}
                )
                SELECT tbl.id
                FROM tbl
                WHERE {conditions}
            )
            RETURNING id
        )
        SELECT COUNT(*) AS deleted_count FROM deleted_rows;
    """

    def __init__(
        self,
        model_enum_value: 'ModelEnumValue',
        *args,
        safe: bool = False,
        log_sql: bool = False,
        **kwargs,
    ):
        self._model_enum_value = model_enum_value
        self._safe = safe
        self._log_sql = log_sql
        self._deleted_count = 0

        super().__init__(*args, **kwargs)

    def get_remove_empty_related_data_conditions(self) -> Optional[str]:
        """Формирование условий очистки данных, содержащих связи с моделями с отсутствующими связанными записями."""
        remove_empty_related_data_conditions = None

        rel_model_condition_template = """
            NOT EXISTS (
                SELECT 1 FROM {0} WHERE {0}.id = tbl.{1}
            )
        """

        rel_model_null_condition_template = """
            tbl.{1} IS NOT NULL AND NOT EXISTS (
                SELECT 1 FROM {0} WHERE {0}.id = tbl.{1}
            )
        """

        remove_empty_related_data_conditions_template = """
            (
                {}
            )
            -- Добавлено условие: удалить только записи старше 30 дней
            AND tbl.modified <= NOW() - INTERVAL '30 days'
        """

        model_label_enum_data = RDMModelEnum.get_model_label_enum_data()
        related_fields = []

        for field in self._model_enum_value.model._meta.concrete_fields:
            if isinstance(field, FictiveForeignKeyField):
                model_label = field.to
            elif isinstance(field, ForeignKey):
                model_label = field.related_model._meta.label
            else:
                continue

            if model_label in model_label_enum_data:
                table_name = model_label_enum_data[model_label].model._meta.db_table
                related_fields.append((field.name, table_name, field.null))

        if related_fields:
            rel_model_conditions_list = []

            for field_name, table_name, is_null in related_fields:
                if is_null:
                    rel_model_conditions_list.append(
                        rel_model_null_condition_template.format(table_name, field_name)
                    )
                else:
                    rel_model_conditions_list.append(rel_model_condition_template.format(table_name, field_name))

            rel_model_conditions = ' OR '.join(rel_model_conditions_list)

            remove_empty_related_data_conditions = remove_empty_related_data_conditions_template.format(
                rel_model_conditions
            )

        return remove_empty_related_data_conditions

    def get_remove_data_by_deleted_records_conditions(self) -> Optional[str]:
        """Формирование условия очистки данных по удаленным записям в ЭШ.

        Если данные в ЭШ были удалены, то после оповещения "Региональной витрины данных" об удалении того или иного
        экземпляра сущности, данные могут быть удалены, т.к. больше не будут обновляться.
        """
        remove_data_by_deleted_records_conditions_sql = None

        if issubclass(self._model_enum_value.model, BaseMainRDMModel):
            remove_data_by_deleted_records_conditions_sql = """
                EXISTS (
                    SELECT 1 
                    FROM rdm_exporting_data_sub_stage redss
                    WHERE redss.id = tbl.exporting_sub_stage_id
                        AND redss.status_id = 'FINISHED'
                        AND tbl.operation = 3                
                )
            """

        return remove_data_by_deleted_records_conditions_sql

    def get_remove_data_by_finished_status_conditions(self) -> Optional[str]:
        """Формирование условия для удаления записей по выгруженным моделям."""
        remove_data_by_finished_status_conditions_sql = None

        if issubclass(self._model_enum_value.model, BaseMainRDMModel):
            remove_data_by_finished_status_conditions_sql = """
                EXISTS (
                    SELECT 1 
                    FROM rdm_exporting_data_sub_stage redss
                    WHERE redss.id = tbl.exporting_sub_stage_id
                        AND redss.status_id = 'FINISHED'
                )
            """

        return remove_data_by_finished_status_conditions_sql

    def get_merged_conditions(self) -> Optional[str]:
        """Формирование общего условия для определения устаревших записей."""
        remove_empty_related_data_conditions = self.get_remove_empty_related_data_conditions()
        remove_data_by_deleted_records_conditions = self.get_remove_data_by_deleted_records_conditions()
        remove_data_by_finished_status_conditions = self.get_remove_data_by_finished_status_conditions()

        main_conditions = []
        if remove_empty_related_data_conditions:
            main_conditions.append(f'({remove_empty_related_data_conditions})')

        if remove_data_by_deleted_records_conditions:
            main_conditions.append(f'({remove_data_by_deleted_records_conditions})')

        if main_conditions:
            conditions = ' OR '.join(main_conditions)
        else:
            return None

        if remove_data_by_finished_status_conditions:
            conditions = f'({conditions}) AND ({remove_data_by_finished_status_conditions})'

        return conditions

    def get_chunk_bounded(self) -> list[tuple[int, int, int]]:
        """Получение идентификаторов-границ чанков для проверки и удаления устаревших данных."""
        get_chunk_bounded_sql = self.SELECT_RDM_CHUNK_BOUNDED_SQL.format(
            chunk_size=settings.RDM_CLEANUP_MODELS_OUTDATED_DATA_CHUNK_SIZE,
            table_name=self._model_enum_value.model._meta.db_table,
        )

        if self._log_sql:
            # Проверка на доступность sqlparse для форматирования
            try:
                import sqlparse
            except ImportError:
                sqlparse = None

            if sqlparse:
                # Форматирование кода
                get_chunk_bounded_sql = sqlparse.format(
                    sql=get_chunk_bounded_sql,
                    reindent=True,
                    strip_comments=True,
                )
            logger.info(
                f'Запрос для получения границ чанков модели {self._model_enum_value.key}: \n{get_chunk_bounded_sql}\n'
            )

        with connection.cursor() as cursor:
            cursor.execute(get_chunk_bounded_sql)
            result = cursor.fetchall()

        return result

    def prepare_queries(self, chunk_bounded: list[tuple[int, int, int]]) -> list[str]:
        """Формирование списка запросов для удаления устаревших данных."""
        queries = []
        conditions = self.get_merged_conditions()

        if conditions:
            for chunk_number, first_id, last_id in chunk_bounded:
                remove_outdated_data_sql = self.REMOVE_OUTDATED_DATA_SQL.format(
                    table_name=self._model_enum_value.model._meta.db_table,
                    first_id=first_id,
                    last_id=last_id,
                    conditions=conditions,
                )

                queries.append(remove_outdated_data_sql)

        return queries

    async def execute_query(self, pool: 'Pool', query: str) -> str:
        """Асинхронное выполнение запроса."""
        async with pool.acquire() as conn:
            try:
                if self._safe:
                    logger.info(f'Запрос не будет выполнен, включен безопасный режим!\n')

                    if self._log_sql:
                        logger.info(f'{query}\n')
                else:
                    deleted_count = await conn.fetchval(query)

                    self._deleted_count += deleted_count

                    if self._log_sql:
                        logger.info(f'При помощи запроса:\n{query}\n')

                    logger.info(f'Было удалено записей: {deleted_count}\n')
            except Exception as e:
                logger.error(f'Ошибка при выполнении {query}\n{e}')

    async def execute_queries(self, queries: list[str]) -> None:
        """Асинхронное выполнение запросов."""
        DB_SETTINGS = settings.DATABASES['default']

        pool = await asyncpg.create_pool(
            max_size=settings.RDM_CLEANUP_MODELS_OUTDATED_DATA_POOL_SIZE,
            min_size=settings.RDM_CLEANUP_MODELS_OUTDATED_DATA_POOL_SIZE,
            host=DB_SETTINGS['HOST'],
            port=DB_SETTINGS['PORT'],
            user=DB_SETTINGS['USER'],
            password=DB_SETTINGS['PASSWORD'],
            database=DB_SETTINGS['NAME'],
        )

        tasks = [self.execute_query(pool, query) for query in queries]

        await asyncio.gather(*tasks)

    def run(self):
        """Запуск очистки устаревших данных."""
        chunk_bounded = self.get_chunk_bounded()

        queries = self.prepare_queries(chunk_bounded=chunk_bounded)

        if queries:
            even_loop = asyncio.new_event_loop()
            try:
                even_loop.run_until_complete(self.execute_queries(queries=queries))
            finally:
                even_loop.close()

        logger.info(f'Удалено записей модели {self._model_enum_value.key}: {self._deleted_count}')



import uuid
from datetime import (
    datetime,
)

from django.db.models import (
    CASCADE,
    PROTECT,
    SET_NULL,
    DateTimeField,
    FileField,
    ForeignKey,
    JSONField,
    PositiveSmallIntegerField,
    UUIDField,
)
from django.utils import (
    timezone,
)
from m3.db import (
    BaseObjectModel,
)

from educommon.django.db.mixins import (
    ReprStrPreModelMixin,
)
from educommon.utils.date import (
    get_today_max_datetime,
    get_today_min_datetime,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledModelEnum,
)

from edu_rdm_integration.core.enums import (
    CommandType,
)
from edu_rdm_integration.core.utils import (
    get_data_command_progress_attachment_path,
)


class RDMCollectingDataStageStatus(TitledModelEnum):
    """Статус этапа сбора данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе сбора',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершено',
    )

    class Meta:
        db_table = 'rdm_collecting_data_stage_status'
        verbose_name = 'Модель-перечисление статусов этапа сбора данных'
        verbose_name_plural = 'Модели-перечисления статусов этапов сбора данных'

    @classmethod
    def get_choices(cls) -> list[tuple[str, str]]:
        """Возвращает список кортежей из ключей и ключей перечисления статусов."""
        return [(key, key) for key in cls.get_model_enum_keys()]


class RDMCollectingDataStage(ReprStrPreModelMixin, BaseObjectModel):
    """Этап подготовки данных в рамках Функций. За работу Функции отвечает ранер менеджер."""

    manager = ForeignKey(
        to='function_tools.Entity',
        verbose_name='Менеджер ранера Функции',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    logs_period_started_at = DateTimeField(
        'Левая граница периода обрабатываемых логов',
        db_index=True,
        default=get_today_min_datetime,
    )

    logs_period_ended_at = DateTimeField(
        'Правая граница периода обрабатываемых логов',
        db_index=True,
        default=get_today_max_datetime,
    )

    started_at = DateTimeField(
        'Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        'Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    status = ForeignKey(
        to='edu_rdm_integration_collect_data_stage.RDMCollectingDataStageStatus',
        verbose_name='Статус',
        on_delete=PROTECT,
        default=RDMCollectingDataStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_collecting_exported_data_stage'
        verbose_name = 'Этап формирования данных для выгрузки'
        verbose_name_plural = 'Этапы формирования данных для выгрузки'

    @property
    def attrs_for_repr_str(self):
        """Список атрибутов для отображения экземпляра модели."""
        return ['manager_id', 'logs_period_started_at', 'logs_period_ended_at', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        """Сохранение этапа сбора данных модели РВД."""
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields'])
            # Добавляем все поля, которые меняются в save()
            kwargs['update_fields'].update({'ended_at'})

        if (
            self.status_id in (RDMCollectingDataStageStatus.FAILED.key, RDMCollectingDataStageStatus.FINISHED.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class RDMCollectingDataSubStageStatus(TitledModelEnum):
    """Статус этапа сбора данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе сбора',
    )

    READY_TO_EXPORT = ModelEnumValue(
        title='Готово к выгрузке',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    EXPORTED = ModelEnumValue(
        title='Выгружено',
    )

    NOT_EXPORTED = ModelEnumValue(
        title='Не выгружено',
    )

    class Meta:
        db_table = 'rdm_collecting_data_sub_stage_status'
        verbose_name = 'Модель-перечисление статусов подэтапа сбора данных'
        verbose_name_plural = 'Модели-перечисления статусов подэтапов сбора данных'


class RDMCollectingDataSubStage(ReprStrPreModelMixin, BaseObjectModel):
    """Подэтап сбора данных для сущностей в рамках функции."""

    stage = ForeignKey(
        to='edu_rdm_integration_collect_data_stage.RDMCollectingDataStage',
        verbose_name='Этап сбора данных',
        on_delete=PROTECT,
    )

    function = ForeignKey(
        to='function_tools.Entity',
        verbose_name='Функция',
        on_delete=PROTECT,
    )

    started_at = DateTimeField(
        'Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        'Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    previous = ForeignKey(
        'self',
        null=True,
        blank=True,
        verbose_name='Предыдущий сбор данных',
        on_delete=CASCADE,
    )

    status = ForeignKey(
        to='edu_rdm_integration_collect_data_stage.RDMCollectingDataSubStageStatus',
        verbose_name='Статус',
        on_delete=PROTECT,
        default=RDMCollectingDataSubStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_collecting_exported_data_sub_stage'
        verbose_name = 'Подэтап сбора данных'
        verbose_name_plural = 'Подэтапы сбора данных'

    @property
    def attrs_for_repr_str(self):
        """Список атрибутов для отображения экземпляра модели."""
        return ['stage_id', 'function_id', 'started_at', 'ended_at', 'previous_id', 'status_id']

    def save(self, *args, **kwargs):
        """Сохранение подэтапа сбора данных."""
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields'])
            # Добавляем все поля, которые меняются в save()
            kwargs['update_fields'].update({'ended_at'})

        if (
            self.status_id
            in (RDMCollectingDataSubStageStatus.FAILED.key, RDMCollectingDataSubStageStatus.READY_TO_EXPORT.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class RDMCollectingDataCommandProgress(ReprStrPreModelMixin, BaseObjectModel):
    """Модель, хранящая данные для формирования и отслеживания асинхронных задач по сбору данных."""

    task = ForeignKey(
        to='async_task.RunningTask',
        verbose_name='Асинхронная задача',
        blank=True,
        null=True,
        on_delete=SET_NULL,
    )
    logs_link = FileField(
        upload_to=get_data_command_progress_attachment_path,
        max_length=255,
        verbose_name='Ссылка на файл логов',
    )
    type = PositiveSmallIntegerField(  # noqa: A003
        verbose_name='Тип команды',
        choices=CommandType.get_choices(),
    )
    stage = ForeignKey(
        to='edu_rdm_integration_collect_data_stage.RDMCollectingDataStage',
        verbose_name='Этап сбора данных',
        blank=True,
        null=True,
        on_delete=SET_NULL,
    )
    model = ForeignKey(
        to='edu_rdm_integration_models.RDMModelEnum',
        verbose_name='Модель РВД',
        on_delete=PROTECT,
    )
    created = DateTimeField(
        verbose_name='Дата создания',
        default=timezone.now,
    )
    logs_period_started_at = DateTimeField(
        'Левая граница периода обрабатываемых логов',
    )
    logs_period_ended_at = DateTimeField(
        'Правая граница периода обрабатываемых логов',
    )
    generation_id = UUIDField(
        'Идентификатор генерации',
        default=uuid.uuid4,
    )
    institute_ids = JSONField(
        'id учебного заведения',
        blank=True,
        null=True,
        default=list,
    )

    class Meta:
        verbose_name = 'Задача по сбору данных'
        verbose_name_plural = 'Задачи по сбору данных'
        db_table = 'edu_rdm_collecting_data_command_progress'



# Generated by Django 3.2.24 on 2025-05-07 19:50

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration', '0016_transferredentity_queue_level'),
    ]

    operations = [
        migrations.DeleteModel(
            name='UploadDataCommand',
        ),
    ]



import inspect
import sys
from importlib import (
    import_module,
)

from educommon.integration_entities.entities import (
    BaseEntity,
)

from edu_rdm_integration.core.utils import (
    register_classes,
)
from edu_rdm_integration.rdm_entities.mixins import (
    EntityEnumRegisterMixin,
)


def is_register_entity(class_) -> bool:
    """Проверяет, является ли класс регистрируемой сущностью."""
    return inspect.isclass(class_) and issubclass(class_, BaseEntity) and issubclass(class_, EntityEnumRegisterMixin)


def register_entities(import_path: str) -> None:
    """Находит регистрируемые сущности в модуле по переданному пути и регистрирует в RegionalDataMartEntityEnum.

    Args:
        import_path: Путь до пакета, хранящего классы сущностей;
    """
    import_module(import_path)
    entities_module = sys.modules[import_path]

    register_classes([c[1] for c in inspect.getmembers(entities_module, is_register_entity)])



# Реализуйте свои unit-тесты здесь
import datetime
import shutil
from typing import (
    TYPE_CHECKING,
)

from django.conf import (
    settings,
)
from django.test import (
    TestCase,
    override_settings,
)
from django.utils import (
    timezone,
)

from edu_function_tools.models import (
    EduEntity,
    EduEntityType,
)

from edu_rdm_integration.core.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
)
from edu_rdm_integration.stages.collect_data.models import (
    CollectingExportedDataStage,
    CollectingExportedDataSubStage,
)
from edu_rdm_integration.stages.export_data.functions.base.consts import (
    TEST_DIR,
)
from edu_rdm_integration.stages.export_data.models import (
    ExportingDataStage,
    ExportingDataSubStage,
)


if TYPE_CHECKING:
    from edu_function_tools.managers import (
        EduRunnerManager,
    )

    from edu_rdm_integration.stages.export_data.functions.base.functions import (
        BaseExportDataFunction,
    )


class BaseExportTestCase(TestCase):
    """Базовый тест экспорта сущности РВД."""

    databases = (settings.DEFAULT_DB_ALIAS, settings.SERVICE_DB_ALIAS)

    def setUp(self) -> None:
        """Подготавливает фикстуры."""
        self.now = timezone.now()

        self.export_period_started_at = datetime.datetime.combine(self.now, datetime.time.min)
        self.export_period_ended_at = datetime.datetime.combine(self.now, datetime.time.max)

    def create_sub_stage(self, *class_names: str) -> CollectingExportedDataSubStage:
        """Создает подэтап сбора данных."""
        function_tools_entities = dict(
            EduEntity.objects.filter(
                type__in=(EduEntityType.MANAGER.key, EduEntityType.FUNCTION.key),
                tags__contains=[REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA],
                class_name__in=class_names,
            ).values_list('type', 'uuid'),
        )
        stage = CollectingExportedDataStage.objects.create(
            manager_id=function_tools_entities[EduEntityType.MANAGER.key],
        )

        return CollectingExportedDataSubStage.objects.create(
            function_id=function_tools_entities[EduEntityType.FUNCTION.key],
            stage=stage,
        )

    @classmethod
    def tearDownClass(cls) -> None:
        """Удаляет временные директории."""
        try:
            shutil.rmtree(TEST_DIR)
        except OSError:
            pass

        super().tearDownClass()


class BaseExportManagerTestCase(BaseExportTestCase):
    """Базовый тест менеджера экспорта сущности РВД."""

    @property
    def manager(self) -> type['EduRunnerManager']:
        """Менеджер раннера Функции экспорта'."""
        raise NotImplementedError

    @override_settings(MEDIA_ROOT=(TEST_DIR + '/media'))
    def run_exporting_manager(self) -> None:
        """Запускает менеджер экспорта."""
        runner_manager = self.manager(
            period_started_at=self.export_period_started_at,
            period_ended_at=self.export_period_ended_at,
        )
        runner_manager.run()

    def get_exporting_stage_sub_stage(self, *class_names) -> tuple[ExportingDataStage, ExportingDataSubStage]:
        """Возвращает этап и подэтап экспорта."""
        function_tools_entities = dict(
            EduEntity.objects.filter(
                type__in=(EduEntityType.MANAGER.key, EduEntityType.FUNCTION.key),
                tags__contains=[REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA],
                class_name__in=class_names,
            ).values_list('type', 'uuid'),
        )
        exported_data_stage = ExportingDataStage.objects.filter(
            manager_id=function_tools_entities[EduEntityType.MANAGER.key],
            period_started_at=self.export_period_started_at,
            period_ended_at=self.export_period_ended_at,
        ).first()
        exported_data_substage = ExportingDataSubStage.objects.filter(
            function_id=function_tools_entities[EduEntityType.FUNCTION.key], stage=exported_data_stage
        ).first()

        return exported_data_stage, exported_data_substage


class BaseExportFunctionTestCase(BaseExportTestCase):
    """Базовый тест менеджера экспорта сущности РВД."""

    @property
    def export_function(self) -> type['BaseExportDataFunction']:
        """Функция экспорта."""
        raise NotImplementedError

    def create_exporting_stage(self, *class_names: str) -> ExportingDataStage:
        """Создает этап экспорта данных."""
        function_tools_entities = dict(
            EduEntity.objects.filter(
                type__in=(EduEntityType.MANAGER.key, EduEntityType.FUNCTION.key),
                tags__contains=[REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA],
                class_name__in=class_names,
            ).values_list('type', 'uuid'),
        )
        return ExportingDataStage.objects.create(
            manager_id=function_tools_entities[EduEntityType.MANAGER.key],
            period_started_at=self.export_period_started_at,
            period_ended_at=self.export_period_ended_at,
        )

    @override_settings(MEDIA_ROOT=(TEST_DIR + '/media'))
    def run_exporting_function(self, exporting_stage: ExportingDataStage, model_ids: list[int]) -> dict:
        """Запускает функцию экспорта."""
        exporting_function = self.export_function(stage=exporting_stage, model_ids=model_ids)
        exporting_function.run()

        return exporting_function.get_function_data()



# Generated by Django 2.2.28 on 2023-12-01 20:58
from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('uploader_client', '0001_initial'),
        ('edu_rdm_integration', '0003_create_index_file_upload_status'),
    ]

    operations = [
        migrations.CreateModel(
            name='UploaderClientLog',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('edu_rdm_integration.exportingdatasubstageuploaderclientlog',),
        ),
    ]



from collections import (
    defaultdict,
)
from typing import (
    TYPE_CHECKING,
)

from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)

from edu_rdm_integration.core.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
)
from edu_rdm_integration.core.storages import (
    RegionalDataMartEntityStorage,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)
from edu_rdm_integration.stages.collect_data.helpers import (
    get_collecting_managers_max_period_ended_dates,
)
from edu_rdm_integration.stages.export_data.helpers import (
    get_exporting_managers_max_period_ended_dates,
)


if TYPE_CHECKING:
    from datetime import (
        datetime,
    )

    from django.db.models import (
        QuerySet,
    )

    from edu_function_tools.managers import (
        EduRunnerManager,
    )

    from edu_rdm_integration.pipelines.transfer.models import (
        TransferredEntity,
    )


class BaseTransferLatestEntitiesDataMixin:
    """Миксин сбора и выгрузки данных."""

    def __init__(self) -> None:
        super().__init__()

        self._collecting_data_managers: dict[str, type['EduRunnerManager']] = {}
        self._collecting_data_manager_to_logs_period_end: dict[str, 'datetime'] = {}

        self._exporting_data_managers: dict[str, type['EduRunnerManager']] = {}
        self._exporting_data_manager_to_period_end: dict[str, 'datetime'] = {}

        self._transferred_entities = []
        self._entites_models_map = defaultdict(list)

    def get_entity_qs(self) -> 'QuerySet[TransferredEntity]':
        """Возвращает QuerySet сущностей сбора и выгрузки."""
        raise NotImplementedError

    def _collect_transferred_entities(self) -> None:
        """Собирает сущности РВД, по которым будет произведен сбор и экспорт данных."""
        self._transferred_entities = [
            (RDMEntityEnum.get_model_enum_value(key=entity), export_enabled)
            for entity, export_enabled in self.get_entity_qs().values_list('entity', 'export_enabled')
        ]

        # Собираем словарь по сущностям с моделями для сборки
        for entity, _ in self._transferred_entities:
            self._entites_models_map[entity.key].extend(
                (
                    model_enum
                    for model_enum in (*entity.additional_model_enums, entity.main_model_enum)
                    if model_enum.order_number != DEFAULT_ORDER_NUMBER
                )
            )

    def _collect_managers(self) -> None:
        """Собирает менеджеры Функций для сбора и выгрузки данных."""
        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        collecting_models_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )
        exporting_entities_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )

        for entity_key, entity_models in self._entites_models_map.items():
            for entity_model in entity_models:
                collect_manager_class = collecting_models_data_managers_map.get(entity_model.key)
                if collect_manager_class:
                    self._collecting_data_managers[entity_model.key] = collect_manager_class

            export_manager_class = exporting_entities_data_managers_map.get(entity_key)
            if export_manager_class:
                self._exporting_data_managers[entity_key] = export_manager_class

    def _calculate_collecting_managers_logs_period_ended_at(self) -> None:
        """Определяет дату последнего успешного этапа сбора у менеджеров Функций сбора."""
        self._collecting_data_manager_to_logs_period_end = get_collecting_managers_max_period_ended_dates(
            self._collecting_data_managers.values()
        )

    def _calculate_exporting_managers_ended_at(self) -> None:
        """Определяет дату последнего успешного подэтапа экспорта у менеджеров Функций экспорта."""
        self._exporting_data_manager_to_period_end = get_exporting_managers_max_period_ended_dates(
            self._exporting_data_managers.values()
        )

    def prepare_collect_export_managers(self) -> None:
        """Подготовка менджеров сбора и экспорта."""
        self._collect_transferred_entities()
        self._collect_managers()
        self._calculate_collecting_managers_logs_period_ended_at()
        self._calculate_exporting_managers_ended_at()



from edu_function_tools.results import (
    EduFunctionResult,
    EduRunnerResult,
)


class BaseCollectingExportedDataRunnerResult(EduRunnerResult):
    """Базовый класс результатов работы ранеров функций сбора данных для интеграции с "Региональная витрина данных"."""


class BaseCollectingExportedDataFunctionResult(EduFunctionResult):
    """Базовый класс результатов работы функций сбора данных для интеграции с "Региональная витрина данных"."""



import os
from collections import (
    defaultdict,
)
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Iterable,
)

from django.conf import (
    settings,
)
from django.db.models import (
    F,
    Model,
    Value,
)
from django.db.models.base import (
    ModelBase,
)
from django.db.models.functions import (
    Concat,
    Substr,
)
from django.db.transaction import (
    atomic,
)
from django.utils import (
    timezone,
)
from django.utils.datastructures import (
    OrderedSet,
)

from educommon import (
    logger,
)
from educommon.async_task.models import (
    RunningTask,
)
from educommon.utils.date import (
    get_today_max_datetime,
)
from educommon.utils.seqtools import (
    make_chunks,
)
from m3_db_utils.consts import (
    DEFAULT_ORDER_NUMBER,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.core.consts import (
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
)
from edu_rdm_integration.core.operations import (
    BaseOperationData,
)
from edu_rdm_integration.core.signals import (
    manager_created,
)
from edu_rdm_integration.core.storages import (
    RegionalDataMartEntityStorage,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)
from edu_rdm_integration.stages.export_data.helpers import (
    get_exporting_managers_max_period_ended_dates,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataStage,
    RDMExportingDataSubStageStatus,
)


if TYPE_CHECKING:
    from edu_function_tools.managers import (
        EduRunnerManager,
    )


class BaseExportEntitiesData(BaseOperationData):
    """Базовый класс экспорта сущностей РВД за указанных период."""

    def __init__(
        self,
        entities: Iterable[str],
        period_started_at=datetime.combine(date.today(), time.min),
        period_ended_at=datetime.combine(date.today(), time.min),
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Если сущности не указаны, берется значение по умолчанию - все сущности:
        entities = entities if entities else RDMEntityEnum.get_enum_data().keys()
        self.entities: list[ModelEnumValue] = [RDMEntityEnum.get_model_enum_value(entity) for entity in entities]

        self.period_started_at = period_started_at
        self.period_ended_at = period_ended_at

        # Классы менеджеров Функций, которые должны быть запущены для выгрузки данных
        self._exporting_data_managers: set[type['EduRunnerManager']] = set()

        # Результаты работы Функций выгрузки данных
        self._exporting_data_results = []

        # Карта соответствия manager_id сущности и его основной модели
        self.manager_main_model_map: dict[str, ModelBase] = {}

    @property
    def _log_file_path(self) -> str:
        """Путь до лог файла."""
        return os.path.join(settings.MEDIA_ROOT, settings.RDM_EXPORT_LOG_DIR, f'{self.command_id}.log')

    def _has_stage_created_or_in_progress(self, manager_id: str, entity: str) -> bool:
        """Проверяет есть ли готовый к работе stage или в работе для данной сущности."""
        stage_created_or_in_progress = RDMExportingDataStage.objects.filter(
            manager_id=manager_id,
            status_id__in=(RDMExportingDataSubStageStatus.CREATED.key, RDMExportingDataSubStageStatus.IN_PROGRESS.key),
        ).exists()

        if stage_created_or_in_progress:
            logger.info(f'entity {entity} is skipped because it is already created or in progress!')

        return stage_created_or_in_progress

    def _fill_manager_entities_map(self, entity_storage: RegionalDataMartEntityStorage) -> None:
        """Заполнение словаря данных с классами менеджеров и их сущностями."""

    def _find_exporting_entities_data_managers(self):
        """Поиск менеджеров Функций выгрузки данных по сущностям РВД."""
        logger.info('find exporting entities data manager..')

        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        exporting_entities_data_managers_map = entity_storage.prepare_entities_manager_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )
        self._fill_manager_entities_map(entity_storage)

        entities = filter(lambda entity: entity.order_number != DEFAULT_ORDER_NUMBER, self.entities)

        for entity_enum in entities:
            manager_class = exporting_entities_data_managers_map.get(entity_enum.key)

            if manager_class and not self._has_stage_created_or_in_progress(manager_class.uuid, entity_enum.key):
                self.manager_main_model_map[manager_class.uuid] = entity_enum.main_model_enum.model
                self._exporting_data_managers.add(manager_class)

        logger.info('finding exporting entities data manager finished.')

    def _export_entities_data(self, *args, **kwargs):
        """Выгрузка данных по указанным сущностям."""
        logger.info('start exporting entities data..')

        kwargs['period_started_at'] = self.period_started_at
        kwargs['period_ended_at'] = self.period_ended_at

        for manager_class in self._exporting_data_managers:
            manager = manager_class(*args, is_only_main_model=True, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager, command_id=self.command_id)

            manager.run()

            self._exporting_data_results.append(manager.result)

        logger.info('exporting entities data finished.')

    def _export(self, *args, **kwargs):
        """Выполнение действий команды."""
        logger.info(f'start exporting data of entities - {", ".join([entity.key for entity in self.entities])}..')

        self._find_exporting_entities_data_managers()
        self._export_entities_data(*args, **kwargs)

        logger.info('exporting entities data finished.')

    def export(self, *args, **kwargs):
        """Запускает экспорт данных."""
        try:
            self._export(*args, **kwargs)
        except Exception as err:
            logger.exception(err)
            raise err
        finally:
            self._remove_file_handler()


class BaseExportLatestEntitiesData(BaseExportEntitiesData):
    """Базовый класс выгрузки сущностей с момента последней успешной выгрузки."""

    def __init__(
        self,
        entities: Iterable[str],
        period_started_at=datetime.combine(date.today(), time.min),
        period_ended_at=datetime.combine(date.today(), time.min),
        update_modified: bool = True,
        **kwargs,
    ):
        super().__init__(entities, period_started_at, period_ended_at, **kwargs)

        self._exporting_data_managers: set[type['EduRunnerManager']] = OrderedSet()

        # Словарь данных с классами менеджеров и их сущностями
        self._manager_entities_map: dict[type[object], list[str]] = defaultdict(set)

        self.async_task = self._get_async_task()
        self.task_id = kwargs.get('task_id')

        self.update_modified = update_modified

    def _get_async_task(self) -> Model:
        """Возвращает модель асинхронной задачи."""
        raise NotImplementedError

    def _set_description_to_async_task(self, exported_entities: Iterable[str]) -> None:
        """Добавляет в описание асинхронной задачи список выгруженных сущностей."""
        if exported_entities and self.task_id:
            entities_str = ', '.join(exported_entities)

            self.async_task.objects.filter(pk=self.task_id).update(
                description=Substr(
                    Concat('description', Value(f': {entities_str}')),
                    1,
                    self.async_task._meta.get_field('description').max_length,
                )
            )

    def _fill_manager_entities_map(self, entity_storage: RegionalDataMartEntityStorage) -> None:
        """Заполнение словаря данных с классами менеджеров и их сущностями."""
        self._manager_entities_map = entity_storage.prepare_manager_entities_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )

    def _update_model_modified_field(self, manager_id: str, last_finished_export_data: datetime) -> None:
        """Обновляет поле modified у невыгруженных записей."""
        if not self.update_modified:
            return

        model = self.manager_main_model_map[manager_id]
        now = timezone.now()

        querysets_to_update = (
            # Не заполнен подэтап выгрузки и modified записи модели < левой границы периода команды latest:
            model.objects.filter(
                exporting_sub_stage_id__isnull=True,
                modified__lt=last_finished_export_data,
            ),
            # Подэтап выгрузки указанной в записи модели имеет статус FAILED
            # и modified записи > даты выгрузки указанной в записи модели сбора (ended_at подэтапа выгрузки)
            # и modified записи модели < левой границы периода команды latest (даты последней удачной выгрузки):
            model.objects.filter(
                exporting_sub_stage__ended_at__gt=now - timedelta(days=365),
                exporting_sub_stage__status_id=RDMExportingDataSubStageStatus.FAILED.key,
                modified__gt=F('exporting_sub_stage__ended_at'),
                modified__lt=last_finished_export_data,
            ),
        )

        for queryset in querysets_to_update:
            not_exported_model_ids = queryset.values_list('id', flat=True).iterator()

            with atomic():
                for model_ids in make_chunks(
                    iterable=not_exported_model_ids,
                    size=settings.RDM_UPDATE_NON_EXPORTED_CHUNK_SIZE,
                ):
                    queryset.filter(id__in=model_ids).update(modified=now)

    def _export_entities_data(self, *args, **kwargs) -> None:
        """Запуск Функций по для экспорта данных."""
        logger.info('export entities data..')

        # Массив с выгружаемыми сущностями для поля "Описание" в асинхронной задаче
        exported_entities = []

        managers_max_period_ended = get_exporting_managers_max_period_ended_dates(self._exporting_data_managers)

        for manager_class in self._exporting_data_managers:
            manager_last_exported = managers_max_period_ended.get(manager_class.uuid)

            kwargs['period_started_at'] = manager_last_exported or timezone.now()
            kwargs['period_ended_at'] = get_today_max_datetime()

            # Обновить поля modified у модели сущности:
            self._update_model_modified_field(
                manager_id=manager_class.uuid,
                last_finished_export_data=kwargs['period_started_at'],
            )

            manager = manager_class(*args, **kwargs)

            if self.command_id:
                # Подается сигнал, что менеджер создан:
                manager_created.send(sender=manager, command_id=self.command_id)

            manager.run()

            self._exporting_data_results.append(manager.result)

            # Если сущность была выгружена, то добавим ее в список exported_entities
            if manager.result.entities and self.task_id:
                exported_entities.extend(self._manager_entities_map.get(manager_class))

        self._set_description_to_async_task(exported_entities)

        logger.info('collecting entities data finished.')


class ExportEntitiesData(BaseExportEntitiesData):
    """Экспорт сущностей РВД за указанных период."""


class ExportLatestEntitiesData(BaseExportLatestEntitiesData):
    """Класс выгрузки сущностей с момента последней успешной выгрузки."""

    def _get_async_task(self) -> Model:
        """Возвращает модель асинхронной задачи."""
        return RunningTask



from abc import (
    ABCMeta,
)

from edu_rdm_integration.stages.collect_data.functions.base.functions import (
    BaseCollectingCalculatedDataFunction,
)
from edu_rdm_integration.stages.collect_data.functions.calculated.base.helpers import (
    BaseCollectingCalculatedExportedDataFunctionHelper,
)
from edu_rdm_integration.stages.collect_data.functions.calculated.base.results import (
    BaseCollectingCalculatedExportedDataFunctionResult,
)
from edu_rdm_integration.stages.collect_data.functions.calculated.base.validators import (
    BaseCollectingCalculatedExportedDataFunctionValidator,
)


class BaseCollectingCalculatedExportedDataFunction(BaseCollectingCalculatedDataFunction, metaclass=ABCMeta):
    """Базовый класс функций сбора расчетных данных для интеграции с "Региональная витрина данных"."""

    def _prepare_helper_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionHelper]:
        """Возвращает класс помощника функции."""
        return BaseCollectingCalculatedExportedDataFunctionHelper

    def _prepare_validator_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionValidator]:
        """Возвращает класс валидатора функции."""
        return BaseCollectingCalculatedExportedDataFunctionValidator

    def _prepare_result_class(self) -> type[BaseCollectingCalculatedExportedDataFunctionResult]:
        """Возвращает класс результата функции."""
        return BaseCollectingCalculatedExportedDataFunctionResult



from m3.db import (
    BaseEnumerate,
)


class FileUploadStatusEnum(BaseEnumerate):
    """Действие по отслеживаемым данным."""

    IN_PROGRESS = 1
    FINISHED = 2
    ERROR = 3
    IN_CHECK = 4

    values = {
        IN_PROGRESS: 'В процессе загрузки в витрину',
        FINISHED: 'Загрузка в витрину закончена',
        ERROR: 'Ошибка загрузки в витрину',
        IN_CHECK: 'На проверке'
    }



# Generated by Django 3.2.24 on 2025-07-04 08:25

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration_collect_data_stage', '0003_auto_20250704_0810'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='rdmcollectingdatasubstage',
            options={'verbose_name': 'Подэтап сбора данных', 'verbose_name_plural': 'Подэтапы сбора данных'},
        ),
        migrations.AlterField(
            model_name='rdmcollectingdatacommandprogress',
            name='stage',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to='edu_rdm_integration_collect_data_stage.rdmcollectingdatastage',
                verbose_name='Этап сбора данных',
            ),
        ),
        migrations.AlterField(
            model_name='rdmcollectingdatasubstage',
            name='stage',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to='edu_rdm_integration_collect_data_stage.rdmcollectingdatastage',
                verbose_name='Этап сбора данных',
            ),
        ),
    ]



from typing_extensions import (
    TYPE_CHECKING,
)

from edu_function_tools.caches import (
    EduFunctionCacheStorage,
    EduRunnerCacheStorage,
)


if TYPE_CHECKING:
    from educommon.integration_entities.entities import (
        BaseEntity,
    )


class BaseExportDataRunnerCacheStorage(EduRunnerCacheStorage):
    """Базовый кеш помощников ранеров функций выгрузки данных для интеграции с "Региональная витрина данных"."""


class BaseExportDataFunctionCacheStorage(EduFunctionCacheStorage):
    """Базовый кеш помощников функций выгрузки данных для интеграции с "Региональная витрина данных"."""

    def _prepare_entity_instances(self, model_ids, *args, **kwargs) -> list['BaseEntity']:
        """Формирование списка объектов сущностей для дальнейшей выгрузки.

        Необходимо переопределить логику формирования объектов у потомков.
        """
        return []

    def __init__(self, *args, model_ids, entities, **kwargs):
        super().__init__(*args, **kwargs)

        self._entity_enum_value = entities[0]
        self._model_ids = model_ids

    def _prepare(self, *args, **kwargs):
        """Заполнение кеша данными."""
        self.entity_instances = self._prepare_entity_instances(model_ids=self._model_ids)



from django.apps.config import (
    AppConfig,
)
from django.conf import (
    settings,
)


class EduRDMIntegrationConfig(AppConfig):
    """Интеграция с Региональной витриной данных."""

    name = 'edu_rdm_integration'
    label = 'edu_rdm_integration'

    def __setup_uploader_client(self):
        """Инициализация клиента для взаимодействия с Региональной витриной данных."""
        import uploader_client
        from django.core.cache import (
            DEFAULT_CACHE_ALIAS,
            caches,
        )
        from uploader_client.contrib.rdm.interfaces.configurations import (
            RegionalDataMartEmulationUploaderConfig,
            RegionalDataMartUploaderConfig,
        )

        if settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION:
            uploader_client.set_config(
                RegionalDataMartEmulationUploaderConfig(
                    interface='uploader_client.contrib.rdm.interfaces.rest.OpenAPIInterfaceEmulation',
                    url=settings.RDM_UPLOADER_CLIENT_URL,
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    timeout=1,
                    request_retries=1,
                    file_status=settings.RDM_RESPONSE_FILE_STATUS,
                )
            )
        elif settings.RDM_UPLOADER_CLIENT_USE_PROXY_API:
            uploader_client.set_config(
                RegionalDataMartUploaderConfig(
                    interface='uploader_client.contrib.rdm.interfaces.rest.ProxyAPIInterface',
                    cache=caches[DEFAULT_CACHE_ALIAS],
                    url=settings.RDM_UPLOADER_CLIENT_URL,
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    timeout=settings.RDM_UPLOADER_CLIENT_REQUEST_TIMEOUT,
                    request_retries=settings.RDM_UPLOADER_CLIENT_REQUEST_RETRIES,
                    organization_ogrn=settings.RDM_UPLOADER_CLIENT_ORGANIZATION_OGRN,
                    installation_name=settings.RDM_UPLOADER_CLIENT_INSTALLATION_NAME,
                    installation_id=settings.RDM_UPLOADER_CLIENT_INSTALLATION_ID,
                    username=settings.RDM_UPLOADER_CLIENT_USERNAME,
                    password=settings.RDM_UPLOADER_CLIENT_PASSWORD,
                )
            )
        else:
            uploader_client.set_config(
                RegionalDataMartUploaderConfig(
                    url=settings.RDM_UPLOADER_CLIENT_URL,
                    datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
                    timeout=settings.RDM_UPLOADER_CLIENT_REQUEST_TIMEOUT,
                    request_retries=settings.RDM_UPLOADER_CLIENT_REQUEST_RETRIES,
                )
            )

    def ready(self):
        """Вызывается после инициализации приложения."""
        super().ready()

        # Инициализация клиента загрузчика данных в Витрину
        self.__setup_uploader_client()



# Generated by Django 3.1.14 on 2024-09-05 20:35

import uuid

import django.db.models.deletion
import django.utils.timezone
from django.db import (
    migrations,
    models,
)

import educommon.django.db.mixins

import edu_rdm_integration.core.utils


class Migration(migrations.Migration):
    dependencies = [
        ('async_task', '0002_task_type_and_status_data'),
        ('edu_rdm_integration', '0013_set_attachment_size'),
    ]

    operations = [
        migrations.CreateModel(
            name='UploadDataCommand',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                (
                    'logs_link',
                    models.FileField(
                        max_length=255,
                        upload_to=edu_rdm_integration.core.utils.get_data_command_progress_attachment_path,
                        verbose_name='Ссылка на файл логов',
                    ),
                ),
                (
                    'type',
                    models.PositiveSmallIntegerField(
                        choices=[(1, 'Автоматический'), (2, 'Ручной')], verbose_name='Тип команды'
                    ),
                ),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Дата создания')),
                ('generation_id', models.UUIDField(default=uuid.uuid4, verbose_name='Идентификатор генерации')),
                (
                    'entity',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to='edu_rdm_integration.regionaldatamartentityenum',
                        verbose_name='Сущность РВД',
                    ),
                ),
                (
                    'task',
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to='async_task.runningtask',
                        verbose_name='Асинхронная задача',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Команда отправки данных в витрину',
                'verbose_name_plural': 'Команды отправки данных в витрину',
                'db_table': 'rdm_upload_data_command',
            },
            bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
        ),
    ]



from abc import (
    ABCMeta,
)

from edu_rdm_integration.stages.collect_data.functions.base.managers import (
    BaseCollectingDataRunnerManager,
)
from edu_rdm_integration.stages.collect_data.functions.calculated.base.runners import (
    BaseCollectingCalculatedExportedDataRunner,
)


class BaseCollectingCalculatedExportedDataRunnerManager(BaseCollectingDataRunnerManager, metaclass=ABCMeta):
    """Менеджер ранеров функций сбора расчетных данных для интеграции с "Региональная витрина данных"."""

    @classmethod
    def _prepare_runner_class(cls) -> type[BaseCollectingCalculatedExportedDataRunner]:
        """Возвращает класс ранера."""
        return BaseCollectingCalculatedExportedDataRunner



from typing import (
    TYPE_CHECKING,
)

import uploader_client
from django.core.cache import (
    DEFAULT_CACHE_ALIAS,
    caches,
)
from django.core.management import (
    BaseCommand,
)
from uploader_client.adapters import (
    UploaderAdapter,
)
from uploader_client.contrib.rdm.interfaces.configurations import (
    RegionalDataMartUploaderConfig,
)


if TYPE_CHECKING:
    from uploader_client.interfaces import (
        OpenAPIRequest,
    )


class BaseDatamartClientCommand(BaseCommand):
    """Базовая команда для загрузки данных/получение статуса в РВД с использованием uploader_client."""

    TIMEOUT = 300
    REQUEST_RETRIES = 1

    def add_arguments(self, parser):
        """Добавление параметров."""
        parser.add_argument(
            '--url',
            type=str,
            required=True,
            help='url хоста Datamart Studio',
        )
        parser.add_argument(
            '--datamart_mnemonic',
            type=str,
            required=True,
            help='мнемоника Витрины',
        )
        parser.add_argument(
            '--organization_ogrn',
            type=str,
            required=True,
            help='ОГРН организации, в рамках которой развёрнута Витрина',
        )
        parser.add_argument(
            '--installation_name',
            type=str,
            required=True,
            help='имя инсталляции в целевой Витрине',
        )
        parser.add_argument(
            '--installation_id',
            type=int,
            required=True,
            help='идентификатор инсталляции (присутствует в её названии)',
        )
        parser.add_argument(
            '--username',
            type=str,
            required=True,
            help='имя пользователя IAM',
        )
        parser.add_argument(
            '--password',
            type=str,
            required=True,
            help='пароль пользователя IAM',
        )

    def _configure_agent_client(
        self,
        url,
        datamart_mnemonic,
        organization_ogrn,
        installation_name,
        installation_id,
        username,
        password,
    ):
        """Конфигурирование клиента загрузчика данных в Витрину."""
        uploader_client.set_config(
            RegionalDataMartUploaderConfig(
                interface='uploader_client.contrib.rdm.interfaces.rest.ProxyAPIInterface',
                cache=caches[DEFAULT_CACHE_ALIAS],
                url=url,
                datamart_name=datamart_mnemonic,
                organization_ogrn=organization_ogrn,
                installation_name=installation_name,
                installation_id=installation_id,
                username=username,
                password=password,
                timeout=self.TIMEOUT,
                request_retries=self.REQUEST_RETRIES,
            )
        )

    def _get_request(self, **options) -> 'OpenAPIRequest':
        """Возвращает запрос для отправки в РВД."""
        raise NotImplementedError

    def handle(self, *args, **options):
        """Выполнение действий команды."""
        self._configure_agent_client(
            url=options['url'],
            datamart_mnemonic=options['datamart_mnemonic'],
            organization_ogrn=options['organization_ogrn'],
            installation_name=options['installation_name'],
            installation_id=options['installation_id'],
            username=options['username'],
            password=options['password'],
        )

        request = self._get_request(**options)

        result = UploaderAdapter().send(request)

        if result.error:
            self.stdout.write(self.style.ERROR('ERROR:\n'))
            self.stdout.write(
                f'{result.error}\nREQUEST:\n"{result.log.request}"\n\nRESPONSE:\n"{result.log.response}"\n'
            )
        else:
            self.stdout.write(self.style.SUCCESS('SUCCESS:\n'))
            self.stdout.write(
                f'Response with {result.response.status_code} code and content:\n{result.response.text}\n'
            )



from abc import (
    abstractmethod,
)
from collections import (
    defaultdict,
)
from typing import (
    Union,
)

from edu_function_tools.caches import (
    EduFunctionCacheStorage,
    EduRunnerCacheStorage,
)
from educommon.audit_log.utils import (
    get_model_by_table,
)
from educommon.integration_entities.consts import (
    LOG_OPERATION_MAP,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from educommon.utils.conversion import (
    int_or_none,
)

from edu_rdm_integration.core.mapping import (
    MODEL_FIELDS_LOG_FILTER,
)
from edu_rdm_integration.stages.collect_data.functions.base.caches import (
    LogChange,
)


class BaseCollectingExportedDataRunnerCacheStorage(EduRunnerCacheStorage):
    """Базовый кеш помощников ранеров функций сбора данных для интеграции с "Региональная витрина данных"."""


class BaseCollectingExportedDataFunctionCacheStorage(EduFunctionCacheStorage):
    """Базовый кеш помощников функций сбора данных для интеграции с "Региональная витрина данных"."""

    def __init__(self, raw_logs, is_merge_logs=False, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Необработанные логи как есть.
        self.raw_logs = raw_logs

        # TODO Перенести в базовый класс (https://jira.bars.group/browse/EDUSCHL-19991)
        # Необходимость объединения логов относящихся к одному объекту
        self.is_merge_logs = is_merge_logs

        # Подготовленные логи в виде:
        # {
        #     'person.person': {
        #         48939: [(1, {'surname': 'Иванов', 'snils': '157-283-394 92'...}), (2, {...})],
        #         44281: [(2, {...}), (2, {...}), ...],
        #         12600: [(3, {...})],
        #     },
        #     'schoolchild.schoolchild': {
        #         ...
        #     },
        # }
        self.logs: dict[str, dict[int, list[LogChange]]] = defaultdict(lambda: defaultdict(list))

    @staticmethod
    def _filter_log(model: str, log_change: LogChange) -> bool:
        """Производится проверка изменений на отслеживаемые поля."""
        is_filtered = False

        if model in MODEL_FIELDS_LOG_FILTER[log_change.operation]:
            filter_fields = MODEL_FIELDS_LOG_FILTER[log_change.operation][model]
            if filter_fields:
                # Если заданы конкретные поля, которые должны отслеживать
                for field in log_change.fields:
                    if field in filter_fields:
                        # Достаточно, чтобы хотя бы одно поле попало под фильтр
                        is_filtered = True
                        break
            else:
                # Модель отслеживается, но перечень фильтруемых полей не задан,
                # значит фильтруем все поля модели
                is_filtered = True

        return is_filtered

    @staticmethod
    def _add_id_to_set(ids: set, _id: Union[int, str, None]) -> None:
        """Добавление значения id во множество ids."""
        _id = int_or_none(_id)
        if _id:
            ids.add(_id)

    def _reformat_logs(self):
        """Производится преобразование логов к удобному для работы виду.

        Предполагается вложенные словари. На первом уровне ключом будет название модели, на втором идентификатор записи.
        """
        for log in self.raw_logs:
            model = get_model_by_table(log.table)._meta.label
            operation = LOG_OPERATION_MAP[log.operation]

            if operation in (EntityLogOperation.CREATE, EntityLogOperation.DELETE):
                fields = log.transformed_data
            elif operation == EntityLogOperation.UPDATE:
                fields = log.transformed_changes
            else:
                fields = {}

            # Если данных нет, то LogChange не формируем
            if not fields:
                continue

            log_change = LogChange(
                operation=operation,
                fields=fields,
            )

            if not self._filter_log(model, log_change):
                # Если модель не отслеживается, то запись лога не сохраняем
                continue

            if log_change.operation == EntityLogOperation.DELETE:
                self.logs[model][log.object_id] = [
                    log_change,
                ]
            else:
                if self.is_merge_logs:
                    self._merge_logs(self.logs[model][log.object_id], log_change)
                else:
                    self.logs[model][log.object_id].append(log_change)

    @staticmethod
    def _merge_logs(log_changes: list[LogChange], log_change: LogChange):
        """Производит объединение (накопление) изменений полей в рамках одного объекта."""
        if log_changes:
            log_changes[0].fields.update(log_change.fields)
        else:
            log_changes.append(log_change)

    def _ignore_logs(self):
        """Исключение логов из обработки на основании описанных правил."""

    def _prepare_logs(self):
        """Подготовка логов для дальнейшей работы."""
        self._reformat_logs()
        self._ignore_logs()

    @abstractmethod
    def _collect_product_model_ids(self):
        """Собирает идентификаторы записей моделей продукта с упором на логи."""

    @abstractmethod
    def _prepare_caches(self):
        """Формирование кэшей."""

    def _prepare(self, *args, **kwargs):
        """Запускает формирование кэша помощника Функции."""
        self._prepare_logs()
        self._collect_product_model_ids()
        self._prepare_caches()



NOT_SPECIFIED_QUEUE_TO_MODEL_ERROR = 'Для переданной модели {} не указана очередь для сохранения!'

CACHED_OBJECTS_TYPE_MISMATCH_ERROR = 'Тип сравниваемого объекта не соответствует типу объекта в кеше'



from django.apps import (
    AppConfig,
)


class RDMCollectDataStageAppConfig(AppConfig):
    """Приложение для организации этапа сбора данных."""

    name = 'edu_rdm_integration.stages.collect_data'
    label = 'edu_rdm_integration_collect_data_stage'
    verbose_name = 'Сбор данных'



# Generated by Django 3.2.24 on 2025-06-28 14:21

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)

import educommon.django.db.mixins
import m3_db_utils.mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ('edu_rdm_integration_export_data_stage', '0001_initial'),
        ('uploader_client', '0001_initial'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.CreateModel(
                    name='DataMartRequestStatus',
                    fields=[
                        ('title', models.TextField(verbose_name='расшифровка значения')),
                        ('value', models.IntegerField(verbose_name='значение')),
                        (
                            'key',
                            models.CharField(
                                db_index=True, max_length=512, primary_key=True, serialize=False, verbose_name='ключ'
                            ),
                        ),
                        ('order_number', models.PositiveIntegerField(default=100000, verbose_name='Порядковый номер')),
                    ],
                    options={
                        'verbose_name': 'Статус загрузки данных в Витрину',
                        'verbose_name_plural': 'Статусы загрузки данных в Витрину',
                        'db_table': 'rdm_request_status',
                    },
                    bases=(m3_db_utils.mixins.BaseValueMixin, models.Model),
                ),
                migrations.CreateModel(
                    name='ExportingDataSubStageUploaderClientLog',
                    fields=[
                        (
                            'id',
                            models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
                        ),
                        (
                            'request_id',
                            models.CharField(
                                blank=True, db_index=True, max_length=100, verbose_name='Id запроса загрузки в витрину'
                            ),
                        ),
                        ('is_emulation', models.BooleanField(default=False, verbose_name='Включен режим эмуляции')),
                        (
                            'file_upload_status',
                            models.SmallIntegerField(
                                blank=True,
                                choices=[
                                    (1, 'В процессе загрузки в витрину'),
                                    (2, 'Загрузка в витрину закончена'),
                                    (3, 'Ошибка загрузки в витрину'),
                                ],
                                null=True,
                                verbose_name='Общий статус загрузки файла в витрину',
                            ),
                        ),
                        (
                            'created',
                            models.DateTimeField(
                                auto_now_add=True, db_index=True, null=True, verbose_name='Дата создания'
                            ),
                        ),
                        (
                            'modified',
                            models.DateTimeField(
                                auto_now=True, db_index=True, null=True, verbose_name='Дата изменения'
                            ),
                        ),
                        (
                            'attachment',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to='edu_rdm_integration_export_data_stage.exportingdatasubstageattachment',
                                verbose_name='Прикрепленный файл',
                            ),
                        ),
                        (
                            'entry',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                related_name='uploader_client_log',
                                to='uploader_client.entry',
                                verbose_name='Лог запроса и ответа',
                            ),
                        ),
                        (
                            'sub_stage',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to='edu_rdm_integration_export_data_stage.exportingdatasubstage',
                                verbose_name='Подэтап выгрузки данных',
                            ),
                        ),
                    ],
                    options={
                        'verbose_name': 'Лог запроса подэтапа выгрузки данных',
                        'verbose_name_plural': 'Лог запроса подэтапа выгрузки данных',
                        'db_table': 'rdm_exporting_data_sub_stage_uploader_client_log',
                    },
                    bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
                ),
                migrations.CreateModel(
                    name='UploadStatusRequestLog',
                    fields=[
                        (
                            'id',
                            models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
                        ),
                        (
                            'entry',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                related_name='upload_status_request_log',
                                to='uploader_client.entry',
                                verbose_name='Cвязь запроса статуса с запросом в витрину',
                            ),
                        ),
                        (
                            'request_status',
                            models.ForeignKey(
                                blank=True,
                                null=True,
                                on_delete=django.db.models.deletion.PROTECT,
                                to='edu_rdm_integration_upload_data_stage.datamartrequeststatus',
                                verbose_name='Статус запроса в витрине',
                            ),
                        ),
                        (
                            'upload',
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to='edu_rdm_integration_upload_data_stage.exportingdatasubstageuploaderclientlog',
                                verbose_name='Cвязь запроса статуса с загрузкой файла в витрину',
                            ),
                        ),
                    ],
                    options={
                        'verbose_name': 'Лог запроса подэтапа выгрузки данных',
                        'verbose_name_plural': 'Логи запроса подэтапа выгрузки данных',
                        'db_table': 'rdm_upload_status_request_log',
                    },
                    bases=(educommon.django.db.mixins.ReprStrPreModelMixin, models.Model),
                ),
            ],
            database_operations=[],
        ),
    ]



from typing import (
    Any,
    Dict,
)

from m3_ext.ui.controls.buttons import (
    ExtButton,
)
from m3_ext.ui.fields import (
    ExtComboBox,
    ExtDisplayField,
    ExtNumberField,
    ExtStringField,
)
from m3_ext.ui.fields.simple import (
    ExtDateTimeField,
)
from m3_ext.ui.icons import (
    Icons,
)
from m3_ext.ui.misc import (
    ExtDataStore,
)

from educommon.objectpack.ui import (
    BaseEditWindow,
)
from educommon.utils.ui import (
    append_template_globals,
)

from edu_rdm_integration.core.consts import (
    BATCH_SIZE,
)
from edu_rdm_integration.core.registry.ui import (
    BaseCreateCommandWindow,
    CommandProgressListWindow,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)


class CreateExportCommandWindow(BaseCreateCommandWindow):
    """Окно создания команды экспорта данных сущности РВД."""

    def _init_components(self):
        """Инициализация компонентов."""
        super()._init_components()

        entity = ExtComboBox(
            name='entity_id',
            label='Сущность',
            display_field='entity',
            anchor='100%',
            editable=False,
            trigger_action_all=True,
            allow_blank=False,
        )
        entity.set_store(ExtDataStore((idx, key) for idx, key in enumerate(RDMEntityEnum.get_model_enum_keys())))
        period_started_at = ExtDateTimeField(
            name='period_started_at',
            label='Начало периода',
            anchor='100%',
            allow_blank=False,
            client_id='period_started_at',
        )
        period_ended_at = ExtDateTimeField(
            name='period_ended_at',
            label='Конец периода',
            anchor='100%',
            allow_blank=False,
            client_id='period_ended_at',
        )
        batch_size = ExtNumberField(
            name='batch_size',
            label='Размер чанка',
            allow_blank=False,
            allow_decimals=False,
            allow_negative=False,
            anchor='100%',
            min_value=1,
            value=BATCH_SIZE,
        )

        self.items_ = (
            entity,
            period_started_at,
            period_ended_at,
            batch_size,
        )

    def set_params(self, params: Dict[str, Any]) -> None:
        """Устанавливает параметры окна."""
        super().set_params(params)

        self.template_globals = 'ui-js/create-export-command-win.js'


class DetailExportCommandWindow(BaseEditWindow):
    """Окно просмотра команды экспорта данных сущности РВД."""

    def set_params(self, params: Dict[str, Any]) -> None:
        """Устанавливает параметры окна."""
        super().set_params(params)

        self.height = 'auto'
        self.logs_link_field.value = params.get('log_url', '')

    def _init_components(self) -> None:
        """Инициализирует компоненты окна."""
        super()._init_components()

        self.entity_field = ExtStringField(
            name='entity_id',
            label='Сущность',
            anchor='100%',
        )
        self.created_field = ExtDateTimeField(
            name='created',
            label='Дата создания',
            anchor='100%',
        )
        self.generation_id_field = ExtStringField(
            name='generation_id',
            label='Идентификатор генерации',
            anchor='100%',
        )
        self.task_id_field = ExtStringField(
            name='task_id',
            label='Идентификатор задачи',
            anchor='100%',
        )
        self.status_field = ExtStringField(
            name='stage.status.key',
            label='Статус экспорта',
            anchor='100%',
        )
        self.started_at_field = ExtDateTimeField(
            name='stage.started_at',
            label='Время начала экспорта',
            anchor='100%',
        )
        self.logs_link_field = ExtDisplayField(
            name='log_url',
            label='Ссылка на логи',
            anchor='100%',
        )
        self.period_started_at_field = ExtDateTimeField(
            name='period_started_at',
            label='Начало периода',
            anchor='100%',
        )
        self.period_ended_at_field = ExtDateTimeField(
            name='period_ended_at',
            label='Конец периода',
            anchor='100%',
        )

    def _do_layout(self) -> None:
        """Располагает компоненты окна."""
        super()._do_layout()

        self.form.items.extend(
            (
                self.entity_field,
                self.created_field,
                self.generation_id_field,
                self.task_id_field,
                self.status_field,
                self.started_at_field,
                self.logs_link_field,
                self.period_started_at_field,
                self.period_ended_at_field,
            )
        )


class ExportCommandProgressListWindow(CommandProgressListWindow):
    """Окно отображения реестра Экспорт данных сущностей РВД."""

    def _init_components(self):
        """Инициализирует компоненты окна."""
        super()._init_components()

        self.sub_stage_for_export_button = ExtButton(
            text='Переотправить',
            icon_cls=Icons.ARROW_REFRESH,
            handler='stageForExport',
        )

    def _do_layout(self):
        """Располагает компоненты окна."""
        super()._do_layout()

        self.grid.top_bar.items.append(self.sub_stage_for_export_button)

    def set_params(self, params):
        """Устанавливает параметры окна."""
        super().set_params(params)

        append_template_globals(self, 'ui-js/stage_for_export.js')
        self.sub_stage_for_export_url = params['sub_stage_for_export_url']



import asyncio
from pathlib import (
    Path,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

from asgiref.sync import (
    sync_to_async,
)
from django.conf import (
    settings,
)
from django.db.models import (
    F,
    Q,
    Sum,
)
from django.db.transaction import (
    atomic,
)

from edu_rdm_integration.core.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.core.redis_cache import (
    AbstractCache,
)
from edu_rdm_integration.stages.export_data.consts import (
    ATTACHMENTS_LIMIT,
    TOTAL_ATTACHMENTS_SIZE_KEY,
)
from edu_rdm_integration.stages.export_data.functions.base.consts import (
    OPERATIONS_METHODS_MAP,
    OPERATIONS_URLS_MAP,
)
from edu_rdm_integration.stages.export_data.functions.base.requests import (
    RegionalDataMartEntityRequest,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataSubStage,
    RDMExportingDataSubStageAttachment,
    RDMExportingDataSubStageStatus,
)
from edu_rdm_integration.stages.upload_data.consts import (
    FAILED_STATUSES,
)
from edu_rdm_integration.stages.upload_data.dataclasses import (
    UploadFile,
)
from edu_rdm_integration.stages.upload_data.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.stages.upload_data.helpers import (
    UploadStatusHelper,
)
from edu_rdm_integration.stages.upload_data.models import (
    RDMExportingDataSubStageUploaderClientLog,
    RDMRequestStatus,
    RDMUploadStatusRequestLog,
)
from edu_rdm_integration.stages.upload_data.queues import (
    Queue,
)
from educommon import (
    logger,
)
from uploader_client.adapters import (
    UploaderAdapter,
)


if TYPE_CHECKING:
    from uploader_client.models import (
        Entry,
    )


class ExportQueueSender:
    """Класс отправки данных в очередь РВД."""

    def __init__(self, data_cache: AbstractCache, queue: Queue, limit: Optional[int] = None):
        self.queue = queue
        self.cache = data_cache
        self.queue_total_file_size = self.cache.get(TOTAL_ATTACHMENTS_SIZE_KEY) or 0
        self.queue_is_full = False
        self.limit = limit or ATTACHMENTS_LIMIT

    @staticmethod
    def get_exported_file_size() -> int:
        """Возвращает размер отправленных в РВД файлов.

        При расчете используются только файлы, отмеченные как отправленные, но не обработанные витриной.
        """
        sended_attachment_ids = RDMExportingDataSubStageUploaderClientLog.objects.filter(
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS
        ).values_list('attachment_id', flat=True)
        file_size = RDMExportingDataSubStageAttachment.objects.filter(id__in=sended_attachment_ids).aggregate(
            Sum('attachment_size')
        )['attachment_size__sum']

        return file_size or 0

    def _make_stage_filter(self) -> Q:
        """Формирование фильтра для выборки подэтапов."""
        return Q(status_id=RDMExportingDataSubStageStatus.READY_FOR_EXPORT.key)

    def get_sub_stages_attachments_to_export(self):
        """Выборка готовых к экспорту подэтапов."""
        sub_stage_ids = set(
            RDMExportingDataSubStage.objects.filter(self._make_stage_filter())
            .order_by('started_at')
            .values_list('id', flat=True)[: self.limit]
        )
        RDMExportingDataSubStage.objects.filter(id__in=sub_stage_ids).update(
            status=RDMExportingDataSubStageStatus.IN_EXPORT.key)

        return (
            RDMExportingDataSubStage.objects.filter(id__in=sub_stage_ids)
            .annotate(
                attachment_id=F('rdmexportingdatasubstageattachment__id'),
                attachment_name=F('rdmexportingdatasubstageattachment__attachment'),
                attachment_size=F('rdmexportingdatasubstageattachment__attachment_size'),
                operation=F('rdmexportingdatasubstageattachment__operation'),
                entity=F('rdmexportingdatasubstageentity__entity_id'),
            )
            .order_by('id', 'operation')
            .values(
                'id',
                'attachment_id',
                'attachment_name',
                'attachment_size',
                'operation',
                'entity',
            )
        )

    def set_sub_stage_to_cache(self, sub_stage_id: int, entity_name: str, attachments: list[UploadFile]) -> bool:
        """Помещение подэтапа в очередь вместе с информацией по файлам и обновление счетчика объема файлов."""
        sub_stage_total_size = sum((attachment.attachment_size for attachment in attachments))
        if self.queue_total_file_size + sub_stage_total_size > settings.RDM_UPLOAD_QUEUE_MAX_SIZE:
            return False

        self.queue.enqueue(sub_stage_id, entity_name, attachments)
        self.queue_total_file_size += sub_stage_total_size
        # Обновим размер файлов в кеш
        self.cache.set(
            TOTAL_ATTACHMENTS_SIZE_KEY, self.queue_total_file_size, timeout=settings.RDM_REDIS_CACHE_TIMEOUT_SECONDS
        )

        logger.info(f'{LOGS_DELIMITER * 2}ExportedDataSubStage {sub_stage_id} {entity_name} added to the queue')

        return True

    def run(self):
        """Запуск работы очереди."""
        if not self.queue_total_file_size:
            self.queue_total_file_size = self.get_exported_file_size()
        if self.queue_total_file_size < settings.RDM_UPLOAD_QUEUE_MAX_SIZE:
            stage_files = []
            prev_sub_stage = None
            entity = ''
            # Если размер очереди позволяет - то отправляем все файлы подэтапа в очередь - иначе прерываем процесс
            for stage_attachment in self.get_sub_stages_attachments_to_export():
                if prev_sub_stage != stage_attachment['id']:
                    if stage_files:
                        to_cache = self.set_sub_stage_to_cache(prev_sub_stage, entity, stage_files)

                        stage_files = []

                        if not to_cache:
                            break

                    prev_sub_stage = stage_attachment['id']

                if stage_attachment['attachment_size']:
                    stage_files.append(
                        UploadFile(
                            stage_attachment['attachment_id'],
                            stage_attachment['attachment_name'],
                            stage_attachment['attachment_size'],
                            stage_attachment['operation'],
                        )
                    )
                    entity = stage_attachment['entity']

            # Обновляем общий объем очереди и закидываем последний элемент
            if stage_files:
                self.set_sub_stage_to_cache(prev_sub_stage, entity, stage_files)
        else:
            # Сохраняем объем отправленных файлов в кеш
            self.cache.set(
                TOTAL_ATTACHMENTS_SIZE_KEY, self.queue_total_file_size, timeout=settings.RDM_REDIS_CACHE_TIMEOUT_SECONDS
            )
            self.queue_is_full = True
            logger.warning(f'Total exported file size:  {self.queue_total_file_size} - queue is full!!!')


class WorkerSender:
    """Непосредственная отправка файлов."""

    def __init__(self, queue: Queue):
        self.queue = queue
        self.entities = set()
        self.received_file_size = 0

    async def get_file_upload_status(
        self,
        request_id: str,
        adapter: UploaderAdapter,
    ) -> tuple[Optional[int], Optional[RDMRequestStatus], 'Entry']:
        """Возвращает статус файла в витрине по запросу."""
        file_upload_status = None

        response, log_entry = await UploadStatusHelper.send_upload_status_request(request_id, adapter)
        request_status = None

        if response:
            request_status = RDMRequestStatus.get_values_to_enum_data().get(response.get('code'))

            if not request_status:
                logger.error(
                    'Не удалось определить статус загрузки данных в витрину. Идентификатор загрузки: '
                    f'{request_id}, данные ответа: {response}',
                )

        if request_status in FAILED_STATUSES:
            file_upload_status = FileUploadStatusEnum.ERROR

        elif request_status == RDMRequestStatus.SUCCESSFULLY_PROCESSED:
            file_upload_status = FileUploadStatusEnum.FINISHED

        return file_upload_status, request_status, log_entry

    async def _process_single_file(
        self,
        file: 'UploadFile',
        sub_stage_id: int,
        entity_key: str,
        adapter: UploaderAdapter,
    ) -> tuple[
        Optional[RDMExportingDataSubStageUploaderClientLog],
        Optional[RDMUploadStatusRequestLog],
        dict[str, Any],
    ]:
        if settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION:
            logger.warning(
                f'{LOGS_DELIMITER * 3}ATTENTION!!! REGIONAL DATA MART INTEGRATION REQUEST EMULATION ENABLED!'
            )
        updates = {
            'stage_status': None,
            'file_size': 0,
            'entity_key': None,
            'errors': []
        }
        file_path = Path.joinpath(Path(settings.MEDIA_ROOT), file.attachment_name)
        file_data = None
        try:
            file_data = await sync_to_async(file_path.open('rb').read, thread_sensitive=False)()
        except (OSError, IOError, FileNotFoundError) as error:
            logger.exception(f'Ошибка чтения файла {file_path} - {str(error)}')
            updates['stage_status'] = RDMExportingDataSubStageStatus.FAILED.key
            return None, None, updates

        method = OPERATIONS_METHODS_MAP.get(file.operation)

        request = RegionalDataMartEntityRequest(
            datamart_name=settings.RDM_UPLOADER_CLIENT_DATAMART_NAME,
            table_name=entity_key.lower(),
            method=method,
            operation=OPERATIONS_URLS_MAP.get(file.operation),
            parameters={},
            headers={
                'Content-Type': 'text/csv',
            },
            files=[],
            data=file_data,
        )

        result = await adapter.send(request)

        request_id = result.response.text if not result.error and result.response else ''
        file_upload_status = FileUploadStatusEnum.IN_PROGRESS if request_id else FileUploadStatusEnum.ERROR

        sub_stage_uploader_client_log = RDMExportingDataSubStageUploaderClientLog(
            entry=result.log,
            sub_stage_id=sub_stage_id,
            attachment_id=file.attachment_id,
            request_id=request_id,
            file_upload_status=file_upload_status,
            is_emulation=settings.RDM_UPLOADER_CLIENT_ENABLE_REQUEST_EMULATION,
        )

        upload_status_request_log = None

        if request_id:
            updated_file_upload_status, request_status, log_entry = (
                await self.get_file_upload_status(request_id, adapter)
            )

            upload_status_request_log = RDMUploadStatusRequestLog(
                upload=sub_stage_uploader_client_log,
                entry=log_entry,
                request_status_id=getattr(request_status, 'key', None),
            )

            if updated_file_upload_status and updated_file_upload_status != file_upload_status:
                sub_stage_uploader_client_log.file_upload_status = updated_file_upload_status

            if updated_file_upload_status == FileUploadStatusEnum.FINISHED:
                updates['file_size'] = file.attachment_size

            if updated_file_upload_status == FileUploadStatusEnum.ERROR:
                updates['stage_status'] = RDMExportingDataSubStageStatus.PROCESS_ERROR.key

        if result.error:
            logger.warning(
                f'{result.error}\nrequest - "{result.log.request}"\nresponse - "{result.log.response}"'
            )
            updates['stage_status'] = RDMExportingDataSubStageStatus.FAILED.key
            updates['errors'].append(result.error)
        else:
            logger.info(f'Response with {result.response.status_code} code and content {result.response.text}')
            updates['entity_key'] = entity_key

        return sub_stage_uploader_client_log, upload_status_request_log, updates

    async def _process_sub_stage(
        self,
        sub_stage_id: int,
        entity_key: str,
        upload_files: list['UploadFile'],
        adapter: UploaderAdapter,
    ):
        sub_stage = await sync_to_async(
            RDMExportingDataSubStage.objects.filter(id=sub_stage_id).first
        )()
        if not sub_stage:
            return

        file_tasks = [
            self._process_single_file(file, sub_stage_id, entity_key, adapter)
            for file in upload_files
        ]
        results = await asyncio.gather(*file_tasks)

        result_to_save = []
        upload_status_requests_to_save = []
        final_status = RDMExportingDataSubStageStatus.FINISHED.key
        total_received_size = 0
        updated_entities = set()

        for log, status_log, updates in results:
            if log:
                result_to_save.append(log)
            if status_log:
                upload_status_requests_to_save.append(status_log)

            if updates.get('stage_status') not in (RDMExportingDataSubStageStatus.FINISHED.key, None):
                final_status = updates.get('stage_status')

            total_received_size += updates.get('file_size', 0)
            if updates.get('entity_key'):
                updated_entities.add(updates['entity_key'])

        @sync_to_async
        @atomic
        def save_and_update_db_and_counters():
            RDMExportingDataSubStageUploaderClientLog.objects.bulk_create(result_to_save)
            RDMUploadStatusRequestLog.objects.bulk_create(upload_status_requests_to_save)
            sub_stage.status_id = final_status
            sub_stage.save()

            self.received_file_size += total_received_size
            self.entities.update(updated_entities)

        await save_and_update_db_and_counters()

        await sync_to_async(self.queue.delete_from_queue)(
            sub_stage_id=sub_stage_id, entity_name=entity_key
        )

        logger.info(f'{LOGS_DELIMITER * 3}ExportedDataSubStage {sub_stage_id} {entity_key} sent from the queue')

    async def send_files(self):
        """Отправка файлов."""
        adapter = UploaderAdapter()
        await adapter.init()

        sub_stages_map = self.queue.dequeue()

        sub_stage_tasks = [
            self._process_sub_stage(sub_stage_id, entity_key, upload_files, adapter)
            for (sub_stage_id, entity_key), upload_files in sub_stages_map.items()
        ]

        await asyncio.gather(*sub_stage_tasks)

        await adapter.close()

    def run(self):
        """Запуск воркера отправки."""
        asyncio.run(self.send_files())



# Generated by Django 3.2.24 on 2025-10-08 14:17

import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):

    dependencies = [
        ('uploader_client', '0001_initial'),
        ('edu_rdm_integration_upload_data_stage', '0002_auto_20250704_0810'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rdmexportingdatasubstageuploaderclientlog',
            name='entry',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='uploader_client_log', to='uploader_client.entry', verbose_name='Лог запроса и ответа'),
        ),
        migrations.AlterField(
            model_name='rdmuploadstatusrequestlog',
            name='entry',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='upload_status_request_log', to='uploader_client.entry', verbose_name='Cвязь запроса статуса с запросом в витрину'),
        ),
    ]



from edu_rdm_integration.stages.collect_data.management.base import (
    BaseCollectModelDataCommand,
)
from edu_rdm_integration.stages.collect_data.operations import (
    BaseCollectLatestModelsData,
    BaseCollectModelsData,
)


class Command(BaseCollectModelDataCommand):
    """Команда для сбора на основе логов за период с последней сборки до указанной даты."""

    def _prepare_collect_models_data_class(self, *args, **kwargs) -> BaseCollectModelsData:
        return BaseCollectLatestModelsData(
            models=kwargs.get('models'),
            logs_period_started_at=kwargs.get('logs_period_started_at'),
            logs_period_ended_at=kwargs.get('logs_period_ended_at'),
        )



from edu_function_tools.errors import (
    EduError,
)


class BaseCollectingExportedDataError(EduError):
    """Базовая ошибка функций сбора данных для интеграции с "Региональная витрина данных"."""



from django.apps import (
    AppConfig,
)


class RDMModelsConfigApp(AppConfig):
    """Приложение для работы с моделями РВД."""

    name = 'edu_rdm_integration.rdm_models'
    label = 'edu_rdm_integration_models'
    verbose_name = 'Модели РВД'



import uuid
from datetime import (
    datetime,
)

from django.db.models import (
    CASCADE,
    PROTECT,
    SET_NULL,
    DateTimeField,
    FileField,
    ForeignKey,
    PositiveIntegerField,
    PositiveSmallIntegerField,
    SmallIntegerField,
    UUIDField,
)
from django.utils import (
    timezone,
)
from m3.db import (
    BaseObjectModel,
)

from edu_function_tools.models import (
    EduEntity,
)
from educommon.django.db.mixins import (
    ReprStrPreModelMixin,
)
from educommon.integration_entities.enums import (
    EntityLogOperation,
)
from m3_db_utils.models import (
    ModelEnumValue,
    TitledModelEnum,
)

from edu_rdm_integration.core.enums import (
    CommandType,
)
from edu_rdm_integration.core.utils import (
    get_data_command_progress_attachment_path,
)
from edu_rdm_integration.rdm_entities.models import (
    RDMEntityEnum,
)
from edu_rdm_integration.stages.export_data.helpers import (
    get_exporting_data_stage_attachment_path,
)


class RDMExportingDataStageStatus(TitledModelEnum):
    """Статус этапа выгрузки данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='В процессе',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершено',
    )

    @classmethod
    def get_choices(cls) -> list[tuple[str, str]]:
        """Возвращает список кортежей из ключей и ключей перечисления статусов."""
        return [(key, key) for key in cls.get_model_enum_keys()]

    class Meta:
        db_table = 'rdm_exporting_data_stage_status'
        verbose_name = 'Модель-перечисление статусов этапа выгрузки данных'
        verbose_name_plural = 'Модели-перечисления статусов этапов выгрузки данных'


class RDMExportingDataStage(ReprStrPreModelMixin, BaseObjectModel):
    """Этап выгрузки данных."""

    manager = ForeignKey(
        to=EduEntity,
        verbose_name='Менеджер ранера Функции',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    period_started_at = DateTimeField(
        'Левая граница периода выборки данных для выгрузки',
        db_index=True,
    )

    period_ended_at = DateTimeField(
        'Правая граница периода выборки данных для выгрузки',
        db_index=True,
    )

    started_at = DateTimeField(
        'Время начала выгрузки данных',
        auto_now_add=True,
    )

    ended_at = DateTimeField(
        'Время завершения выгрузки данных',
        null=True,
        blank=True,
    )

    status = ForeignKey(
        to=RDMExportingDataStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=RDMExportingDataStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_exporting_data_stage'
        verbose_name = 'Этап формирования данных для выгрузки'
        verbose_name_plural = 'Этапы формирования данных для выгрузки'

    @property
    def attrs_for_repr_str(self):
        """Список атрибутов для отображения экземпляра модели."""
        return ['manager_id', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        """Сохранение этапа экспорта данных."""
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields'])
            # Добавляем все поля, которые меняются в save()
            kwargs['update_fields'].update({'ended_at'})

        if (
            self.status_id in (RDMExportingDataStageStatus.FAILED.key, RDMExportingDataStageStatus.FINISHED.key)
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class RDMExportingDataSubStageStatus(TitledModelEnum):
    """Модель-перечисление статусов этапа выгрузки данных."""

    CREATED = ModelEnumValue(
        title='Создан',
    )

    IN_PROGRESS = ModelEnumValue(
        title='Запущен',
    )

    FAILED = ModelEnumValue(
        title='Завершено с ошибками',
    )

    FINISHED = ModelEnumValue(
        title='Завершен',
    )
    READY_FOR_EXPORT = ModelEnumValue(
        title='Готов к выгрузке',
    )
    PROCESS_ERROR = ModelEnumValue(title='Ошибка обработки витриной')

    IN_EXPORT =  ModelEnumValue(
        title='Экспортируется',
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_status'
        verbose_name = 'Модель-перечисление статусов подэтапа выгрузки данных'
        verbose_name_plural = 'Модели-перечисления статусов подэтапов выгрузки данных'

    @classmethod
    def get_not_finished_statuses(cls):
        """Получить список статусов, которые не завершены."""
        return cls.CREATED, cls.IN_PROGRESS, cls.FAILED, cls.READY_FOR_EXPORT, cls.PROCESS_ERROR


class RDMExportingDataSubStage(ReprStrPreModelMixin, BaseObjectModel):
    """Подэтап выгрузки данных."""

    function = ForeignKey(
        to=EduEntity,
        verbose_name='Функция',
        on_delete=PROTECT,
        null=True,
        blank=True,
    )

    stage = ForeignKey(
        to=RDMExportingDataStage,
        verbose_name='Этап выгрузки данных',
        on_delete=CASCADE,
    )

    started_at = DateTimeField(
        verbose_name='Время начала сбора данных',
        auto_now_add=True,
        db_index=True,
    )

    ended_at = DateTimeField(
        verbose_name='Время завершения сбора данных',
        null=True,
        blank=True,
        db_index=True,
    )

    status = ForeignKey(
        to=RDMExportingDataSubStageStatus,
        verbose_name='Статус',
        on_delete=PROTECT,
        default=RDMExportingDataSubStageStatus.CREATED.key,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage'
        verbose_name = 'Стадия выгрузки данных'
        verbose_name_plural = 'Стадии выгрузки данных'

    @property
    def attrs_for_repr_str(self):
        """Список атрибутов для отображения экземпляра модели."""
        return ['function_id', 'collecting_data_sub_stage_id', 'stage_id', 'started_at', 'ended_at', 'status_id']

    def save(self, *args, **kwargs):
        """Сохранение экземпляра модели."""
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields'])
            # Добавляем все поля, которые меняются в save()
            kwargs['update_fields'].update({'ended_at'})

        if (
            self.status_id
            in {
                RDMExportingDataSubStageStatus.FAILED.key,
                RDMExportingDataSubStageStatus.FINISHED.key,
                RDMExportingDataSubStageStatus.READY_FOR_EXPORT.key,
            }
            and not self.ended_at
        ):
            self.ended_at = datetime.now()

        super().save(*args, **kwargs)


class RDMExportingDataSubStageAttachment(ReprStrPreModelMixin, BaseObjectModel):
    """Сгенерированный файл для дальнейшей выгрузки в "Региональная витрина данных"."""

    exporting_data_sub_stage = ForeignKey(
        to=RDMExportingDataSubStage,
        verbose_name='Подэтап выгрузки данных',
        null=True,
        blank=True,
        on_delete=SET_NULL,
    )

    # TODO PYTD-22 В зависимости от принятого решения по инструменту ограничения доступа к media-файлам, нужно будет
    #  изменить тип поля или оставить как есть
    attachment = FileField(
        verbose_name='Сгенерированный файл',
        upload_to=get_exporting_data_stage_attachment_path,
        max_length=512,
        null=True,
        blank=True,
    )

    operation = SmallIntegerField(
        verbose_name='Действие',
        choices=EntityLogOperation.get_choices(),
    )

    created = DateTimeField(
        verbose_name='Дата создания',
        auto_now_add=True,
        null=True,
        blank=True,
    )
    modified = DateTimeField(
        verbose_name='Дата изменения',
        auto_now=True,
        null=True,
        blank=True,
    )
    attachment_size = PositiveIntegerField(null=True, verbose_name='Размер файла (байт)')

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_attachment'
        verbose_name = 'Сгенерированный файл для дальнейшей выгрузки в "Региональная витрина данных"'
        verbose_name_plural = 'Сгенерированные файлы для дальнейшей выгрузки в "Региональная витрина данных"'

    @property
    def attrs_for_repr_str(self):
        """Список атрибутов для отображения экземпляра модели."""
        return ['exporting_data_sub_stage_id', 'attachment', 'operation', 'created', 'modified']


class RDMExportingDataSubStageEntity(BaseObjectModel):
    """Модель связи сущности и подэтапа выгрузки."""

    entity = ForeignKey(
        to=RDMEntityEnum,
        verbose_name='Сущность РВД',
        on_delete=PROTECT,
    )

    exporting_data_sub_stage = ForeignKey(
        to=RDMExportingDataSubStage,
        verbose_name='Подэтап выгрузки данных',
        on_delete=CASCADE,
    )

    class Meta:
        db_table = 'rdm_exporting_data_sub_stage_entity'
        verbose_name = 'Связь сущности и подэтапа выгрузки'
        verbose_name_plural = 'Связи сущности и подэтапа выгрузки'


class RDMExportingDataCommandProgress(ReprStrPreModelMixin, BaseObjectModel):
    """Команда экспорта данных."""

    task = ForeignKey(
        to='async_task.RunningTask',
        verbose_name='Асинхронная задача',
        null=True,
        blank=True,
        on_delete=SET_NULL,
    )
    logs_link = FileField(
        upload_to=get_data_command_progress_attachment_path,
        max_length=255,
        verbose_name='Файл лога',
    )
    type = PositiveSmallIntegerField(  # noqa: A003
        verbose_name='Тип команды',
        choices=CommandType.get_choices(),
    )
    stage = ForeignKey(
        to=RDMExportingDataStage,
        verbose_name='Этап выгрузки данных',
        null=True,
        blank=True,
        on_delete=SET_NULL,
    )
    entity = ForeignKey(
        to=RDMEntityEnum,
        verbose_name='Сущность РВД',
        on_delete=PROTECT,
    )
    created = DateTimeField(
        verbose_name='Дата создания',
        default=timezone.now,
    )
    period_started_at = DateTimeField(
        'Левая граница периода выборки данных для выгрузки',
    )
    period_ended_at = DateTimeField(
        'Правая граница периода выборки данных для выгрузки',
    )
    generation_id = UUIDField(
        'Идентификатор генерации',
        default=uuid.uuid4,
    )

    class Meta:
        db_table = 'edu_rdm_exporting_data_command_progress'
        verbose_name = 'Команда экспорта данных'
        verbose_name_plural = 'Команды экспорта данных'



# Generated by Django 3.1.14 on 2024-08-19 14:24

from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration', '0011_exportingdatasubstageentity'),
    ]

    operations = [
        migrations.AddField(
            model_name='exportingdatasubstageattachment',
            name='attachment_size',
            field=models.PositiveIntegerField(null=True, verbose_name='Размер файла (байт)'),
        ),
    ]



from edu_rdm_integration.stages.upload_data.models import (
    RDMExportingDataSubStageUploaderClientLog,
    RDMUploadStatusRequestLog,
)
from uploader_client.models import (
    Entry,
)

from .base import (
    BaseServiceOutdatedDataCleaner,
)
from .export_data import (
    ExportingDataSubStageAttachmentCleaner,
    ExportingDataSubStageCleaner,
)


class ExportingDataSubStageUploaderClientLogCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка логов загрузчика подэтапов выгрузки данных без связи с подэтапами или файлами."""

    model = RDMExportingDataSubStageUploaderClientLog

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        sub_stage_table = ExportingDataSubStageCleaner.get_table_name()
        attachment_table = ExportingDataSubStageAttachmentCleaner.get_table_name()

        conditions = [
            f'({self.get_status_condition(sub_stage_table, "id", "FINISHED", 7, "sub_stage_id")})',
            f'({self.get_status_condition(sub_stage_table, "id", "FAILED", 30, "sub_stage_id")})',
            f'({self.get_orphan_reference_condition(sub_stage_table, "id", local_field="sub_stage_id")})',
            f'({self.get_orphan_reference_condition(attachment_table, "id", local_field="attachment_id")})',
        ]

        return ' OR '.join(conditions)


class UploadStatusRequestLogCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка логов статуса загрузки файла в витрину без связей upload."""

    model = RDMUploadStatusRequestLog

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        uploader_client_log_table = ExportingDataSubStageUploaderClientLogCleaner.get_table_name()

        return self.get_orphan_reference_condition(uploader_client_log_table, 'id', local_field='upload_id')


class EntryCleaner(BaseServiceOutdatedDataCleaner):
    """Очистка записей журнала, не связанные ни с upload, ни с логами."""

    model = Entry

    def get_merged_conditions(self) -> str:
        """Формирует условие удаления для устаревших данных."""
        uploader_client_log_table = ExportingDataSubStageUploaderClientLogCleaner.get_table_name()
        upload_status_log_table = UploadStatusRequestLogCleaner.get_table_name()
        conditions = [
            f'({self.get_orphan_reference_condition(uploader_client_log_table, "entry_id")})',
            f'({self.get_orphan_reference_condition(upload_status_log_table, "entry_id")})',
        ]

        return ' AND '.join(conditions)



# Generated by Django 3.2.24 on 2025-07-04 08:10

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration_export_data_stage', '0002_auto_20250704_0810'),
        ('uploader_client', '0001_initial'),
        ('edu_rdm_integration_upload_data_stage', '0001_initial'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='ExportingDataSubStageUploaderClientLog',
            new_name='RDMExportingDataSubStageUploaderClientLog',
        ),
        migrations.RenameModel(
            old_name='DataMartRequestStatus',
            new_name='RDMRequestStatus',
        ),
        migrations.RenameModel(
            old_name='UploadStatusRequestLog',
            new_name='RDMUploadStatusRequestLog',
        ),
    ]



from collections import (
    defaultdict,
)

from edu_function_tools.models import (
    EduEntityType,
)
from edu_function_tools.runners import (
    EduRunner,
)
from edu_function_tools.storages import (
    EduEntityStorage,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.core.consts import (
    REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA,
    REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA,
)


class RegionalDataMartEntityStorage(EduEntityStorage):
    """Хранилище классов сущностей реализованных в системе.

    Собираются только те сущности, типы которых указаны в модели-перечислении
    edu_function_tools.models.EduEntityType. Расширение произведено в связи с необходимостью получения
    карт соответствия менеджеров и сущностей, функций и сущностей.
    """

    def _collect_runner_regional_data_mart_integration_entities(
        self,
        runner_class: type[EduRunner],
        runner_regional_data_mart_integration_entities: list[str],
    ):
        """Собирает и возвращает список сущностей."""
        for runnable_class in runner_class._prepare_runnable_classes():
            if hasattr(runnable_class, '_prepare_runnable_classes'):
                self._collect_runner_regional_data_mart_integration_entities(
                    runner_class=runnable_class,
                    runner_regional_data_mart_integration_entities=runner_regional_data_mart_integration_entities,
                )

                continue

            if hasattr(runnable_class, 'entities'):
                entities = getattr(runnable_class, 'entities')

                runner_regional_data_mart_integration_entities.extend(entities)

    def prepare_entities_map(
        self,
        entity_type: ModelEnumValue,
        tags: set[str],
        *args,
        **kwargs,
    ) -> dict[str, type[object]]:
        """Формирование карты соответствия сущности интеграции с РВД и сущности function_tools."""
        rdm_integration_function_tools_entities_map: dict[str, type[object]] = {}

        registered_function_tools_entities_classes = [
            entity['class']
            for entity in self.entities[entity_type.key].values()
            if not tags.difference(entity['class'].tags)
        ]

        for function_tools_entity_class in registered_function_tools_entities_classes:
            regional_data_mart_integration_entities = []

            if hasattr(function_tools_entity_class, 'runner_class'):
                runner_class = function_tools_entity_class.runner_class

                self._collect_runner_regional_data_mart_integration_entities(
                    runner_class=runner_class,
                    runner_regional_data_mart_integration_entities=regional_data_mart_integration_entities,
                )
            elif hasattr(function_tools_entity_class, 'entities'):
                regional_data_mart_integration_entities = function_tools_entity_class.entities

            for rdm_integration_entity in regional_data_mart_integration_entities:
                rdm_integration_function_tools_entities_map[rdm_integration_entity.key] = function_tools_entity_class

        return rdm_integration_function_tools_entities_map

    def prepare_entities_manager_map(self, tags: set[str]) -> dict[str, type[object]]:
        """Формирование карты соответствия сущности интеграции с "Региональная витрина данных" и менеджера Функции.

        Стоит учитывать, что в рамках одной Функции может производиться работа с несколькими сущностями.
        """
        return self.prepare_entities_map(
            entity_type=EduEntityType.MANAGER,
            tags=tags,
        )

    def prepare_manager_entities_map(self, tags: set[str]) -> dict[type[object], list[str]]:
        """Формирование карты соответствия менеджера Функции и сущности интеграции с "Региональная витрина данных".

        Стоит учитывать, что в рамках одной Функции может производиться работа с несколькими сущностями.
        """
        _manager_entities_map = defaultdict(set)
        exporting_entities_data_managers_map = self.prepare_entities_manager_map(tags=tags)

        for entity in exporting_entities_data_managers_map:
            _manager_entities_map[exporting_entities_data_managers_map[entity]].add(entity)

        return _manager_entities_map

    def prepare_entities_functions_map(self, tags: set[str]) -> dict[str, type[object]]:
        """Формирование карты соответствия сущности интеграции с "Региональная витрина данных" и функции.

        Стоит учитывать, что в рамках одной Функции может производиться работа с несколькими сущностями.
        """
        return self.prepare_entities_map(
            entity_type=EduEntityType.FUNCTION,
            tags=tags,
        )

    def prepare_exporting_collecting_functions_map(self) -> dict[str, str]:
        """Возвращает карту соответствия функций выгрузки и сбора данных.

        Returns:
            Возвращает словарь в качестве ключей и значений используются UUID-ы функций.
        """
        exporting_collecting_functions_map = {}

        exporting_entities_data_functions_map = self.prepare_entities_functions_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_EXPORTING_DATA},
        )
        collecting_entities_data_functions_map = self.prepare_entities_functions_map(
            tags={REGIONAL_DATA_MART_INTEGRATION_COLLECTING_DATA},
        )

        for entity in exporting_entities_data_functions_map.keys():
            exporting_collecting_functions_map[exporting_entities_data_functions_map[entity].uuid] = (
                collecting_entities_data_functions_map[entity].uuid
            )

        return exporting_collecting_functions_map



from django.apps import (
    AppConfig,
)


class RDMEntitiesAppConfig(AppConfig):
    """Приложение для работы с сущностями РВД."""

    name = 'edu_rdm_integration.rdm_entities'
    label = 'edu_rdm_integration_entities'
    verbose_name = 'Сущности РВД'



import django.db.models.deletion
from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    """Миграция."""

    dependencies = [
        ('edu_rdm_integration', '0007_delete_upload_status'),
    ]

    operations = [
        migrations.CreateModel(
            name='TransferredEntity',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'entity',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='edu_rdm_integration.RegionalDataMartEntityEnum',
                        verbose_name='Сущность',
                    ),
                ),
            ],
            options={
                'verbose_name': 'Сущность, по которой должен быть произведен сбор и экспорт данных',
                'verbose_name_plural': 'Сущности, по которым должен быть произведен сбор и экспорт данных',
                'db_table': 'rdm_transferred_entity',
            },
        ),
    ]



import celery
from celery.schedules import (
    crontab,
)
from django.conf import (
    settings,
)
from django.core.cache import (
    cache,
)

from edu_rdm_integration.core.consts import (
    TASK_QUEUE_NAME,
)
from edu_rdm_integration.stages.upload_data.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.stages.upload_data.helpers import (
    UploadStatusHelper,
)
from edu_rdm_integration.stages.upload_data.models import (
    RDMExportingDataSubStageUploaderClientLog,
)
from edu_rdm_integration.stages.upload_data.operations import (
    UploadData,
)
from edu_rdm_integration.stages.upload_data.queues import (
    RdmDictBasedSubStageAttachmentQueue,
)
from educommon.async_task.models import (
    AsyncTaskType,
)
from educommon.async_task.tasks import (
    SelfReschedulingAsyncTask,
)


class RDMCheckUploadStatus(SelfReschedulingAsyncTask):
    """Периодическая задача для сбора статусов по загрузке файла в витрину."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Сбор статусов загрузки данных в витрину "Региональная витрина данных"'
    task_type = AsyncTaskType.UNKNOWN
    next_task_delay_seconds = settings.RDM_UPLOAD_DATA_NEXT_TASK_DELAY_SECONDS
    lock_expire_seconds = settings.RDM_UPLOAD_DATA_TASK_LOCK_EXPIRE_SECONDS
    run_every = crontab(
        minute=settings.RDM_UPLOAD_DATA_TASK_MINUTE,
        hour=settings.RDM_UPLOAD_DATA_TASK_HOUR,
        day_of_week=settings.RDM_UPLOAD_DATA_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение."""
        super().process(*args, **kwargs)

        in_progress_uploads = RDMExportingDataSubStageUploaderClientLog.objects.filter(
            file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
            is_emulation=False,
        ).select_related('attachment')

        UploadStatusHelper(in_progress_uploads, cache).run()


class UploadDataAsyncTask(SelfReschedulingAsyncTask):
    """Формирование очереди файлов и их отправка."""

    queue = TASK_QUEUE_NAME
    routing_key = TASK_QUEUE_NAME
    description = 'Отправка данных в витрину "Региональная витрина данных"'
    task_type = AsyncTaskType.SYSTEM
    next_task_delay_seconds = settings.RDM_UPLOAD_DATA_NEXT_TASK_DELAY_SECONDS
    lock_expire_seconds = settings.RDM_UPLOAD_DATA_TASK_LOCK_EXPIRE_SECONDS
    run_every = crontab(
        minute=settings.RDM_UPLOAD_DATA_TASK_MINUTE,
        hour=settings.RDM_UPLOAD_DATA_TASK_HOUR,
        day_of_week=settings.RDM_UPLOAD_DATA_TASK_DAY_OF_WEEK,
    )

    def process(self, *args, **kwargs):
        """Выполнение."""
        super().process(*args, **kwargs)

        queue = RdmDictBasedSubStageAttachmentQueue()
        upload_data = UploadData(
            data_cache=cache,
            queue=queue,
        )

        upload_result = upload_data.upload_data()

        task_result = {
            'Общий объем отправленных файлов': f'{upload_result["total_file_size"]}',
            'Очередь отправки переполнена': 'Да' if upload_result['queue_is_full'] else 'Нет',
            'Сущности, отправленные в витрину': upload_result['uploaded_entities'],
        }

        self.set_progress(values=task_result)


celery_app = celery.app.app_or_default()
celery_app.register_task(RDMCheckUploadStatus)
celery_app.register_task(UploadDataAsyncTask)



from time import (
    sleep,
)
from typing import (
    TYPE_CHECKING,
    Any,
)

from django.core.cache import (
    cache,
)
from django.core.management.base import (
    BaseCommand,
)

from edu_rdm_integration.core.consts import (
    BATCH_SIZE,
)
from edu_rdm_integration.stages.upload_data.enums import (
    FileUploadStatusEnum,
)
from edu_rdm_integration.stages.upload_data.helpers import (
    UploadStatusHelper,
)
from edu_rdm_integration.stages.upload_data.models import (
    RDMExportingDataSubStageUploaderClientLog,
)


class Command(BaseCommand):
    """Команда для отправки данных в витрину параллельно-последовательно. В рамках скрипта последовательно,
    параллельно количетвом запусков команды."""

    help = 'Команда для отправки данных в витрину параллельно-последовательно'  # noqa: A003


    def handle(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> None:
        """Обработчик команды."""
        while True:
            self.stdout.write(f'Начало проверки статуса загрузки данных в витрину..')

            # Получаем незавершенные загрузки данных в витрину
            in_progress_uploads = RDMExportingDataSubStageUploaderClientLog.objects.filter(
                file_upload_status=FileUploadStatusEnum.IN_PROGRESS,
                is_emulation=False,
            ).select_related('attachment')[:BATCH_SIZE]

            for upload in in_progress_uploads:
                upload.file_upload_status = FileUploadStatusEnum.IN_CHECK

            RDMExportingDataSubStageUploaderClientLog.objects.bulk_update(in_progress_uploads, fields=['file_upload_status'])

            self.stdout.write(f'Обновление статуса загрузки данных в витрину на {FileUploadStatusEnum.IN_CHECK}..')

            UploadStatusHelper(in_progress_uploads, cache).run()

            sleep(10)

            self.stdout.write(f'Окончание проверки статуса загрузки данных в витрину.\n\n')



from educommon.integration_entities.enums import (
    EntityLogOperation,
)


POST = 'POST'
# удаление еще не реализовано, ожидается информация от МинЦифры
DELETE = 'DELETE'

OPERATIONS_METHODS_MAP = {
    EntityLogOperation.CREATE: POST,
    EntityLogOperation.UPDATE: POST,
    EntityLogOperation.DELETE: POST,
}

OPERATIONS_URLS_MAP = {
    EntityLogOperation.CREATE: 'upload',
    EntityLogOperation.UPDATE: 'upload',
    EntityLogOperation.DELETE: 'delete',
}


TEST_DIR = 'test_data'



# Generated by Django 3.2.24 on 2025-07-04 07:25

from django.db import (
    migrations,
)


class Migration(migrations.Migration):
    dependencies = [
        ('rdm_collect_and_export_data', '0003_auto_20250704_0725'),
        ('edu_rdm_integration', '0017_delete_uploaddatacommand'),
    ]

    operations = [
        migrations.SeparateDatabaseAndState(
            state_operations=[
                migrations.RemoveField(
                    model_name='collectingexporteddatastage',
                    name='manager',
                ),
                migrations.RemoveField(
                    model_name='collectingexporteddatastage',
                    name='status',
                ),
                migrations.RemoveField(
                    model_name='collectingexporteddatasubstage',
                    name='function',
                ),
                migrations.RemoveField(
                    model_name='collectingexporteddatasubstage',
                    name='previous',
                ),
                migrations.RemoveField(
                    model_name='collectingexporteddatasubstage',
                    name='stage',
                ),
                migrations.RemoveField(
                    model_name='collectingexporteddatasubstage',
                    name='status',
                ),
                migrations.RemoveField(
                    model_name='exportingdatastage',
                    name='manager',
                ),
                migrations.RemoveField(
                    model_name='exportingdatastage',
                    name='status',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstage',
                    name='function',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstage',
                    name='stage',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstage',
                    name='status',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageattachment',
                    name='exporting_data_sub_stage',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageentity',
                    name='entity',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageentity',
                    name='exporting_data_sub_stage',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageuploaderclientlog',
                    name='attachment',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageuploaderclientlog',
                    name='entry',
                ),
                migrations.RemoveField(
                    model_name='exportingdatasubstageuploaderclientlog',
                    name='sub_stage',
                ),
                migrations.DeleteModel(
                    name='RegionalDataMartModelEnum',
                ),
                migrations.RemoveField(
                    model_name='transferredentity',
                    name='entity',
                ),
                migrations.RemoveField(
                    model_name='uploadstatusrequestlog',
                    name='entry',
                ),
                migrations.RemoveField(
                    model_name='uploadstatusrequestlog',
                    name='request_status',
                ),
                migrations.RemoveField(
                    model_name='uploadstatusrequestlog',
                    name='upload',
                ),
                migrations.DeleteModel(
                    name='UploaderClientLog',
                ),
                migrations.DeleteModel(
                    name='CollectingDataStageStatus',
                ),
                migrations.DeleteModel(
                    name='CollectingDataSubStageStatus',
                ),
                migrations.DeleteModel(
                    name='CollectingExportedDataStage',
                ),
                migrations.DeleteModel(
                    name='CollectingExportedDataSubStage',
                ),
                migrations.DeleteModel(
                    name='DataMartRequestStatus',
                ),
                migrations.DeleteModel(
                    name='ExportingDataStage',
                ),
                migrations.DeleteModel(
                    name='ExportingDataStageStatus',
                ),
                migrations.DeleteModel(
                    name='ExportingDataSubStage',
                ),
                migrations.DeleteModel(
                    name='ExportingDataSubStageAttachment',
                ),
                migrations.DeleteModel(
                    name='ExportingDataSubStageEntity',
                ),
                migrations.DeleteModel(
                    name='ExportingDataSubStageStatus',
                ),
                migrations.DeleteModel(
                    name='ExportingDataSubStageUploaderClientLog',
                ),
                migrations.DeleteModel(
                    name='RegionalDataMartEntityEnum',
                ),
                migrations.DeleteModel(
                    name='TransferredEntity',
                ),
                migrations.DeleteModel(
                    name='UploadStatusRequestLog',
                ),
            ],
        ),
    ]



from django.apps import (
    AppConfig as AppConfigBase,
)


class AppConfig(AppConfigBase):
    """Базовые компоненты для запуска сбора и выгрузки данных через пользовательский интерфейс."""

    name = __package__
    label = 'rdm_collect_and_export_data'



from django.db import (
    connection,
)


def update_foreign_key_constraint(
    table_name: str,
    field_name: str,
    target_table: str,
    on_delete: str,
):
    """
    Обновляет внешний ключ (FOREIGN KEY) в PostgreSQL.

    Выполняет:
      1. Поиск существующего constraint'а по таблице и целевой модели.
      2. Удаление найденного constraint'а.
      3. Создание нового constraint'а с заданным ON DELETE.

    Args:
        table_name (str): Имя таблицы, содержащей внешний ключ.
        field_name (str): Имя поля внешнего ключа.
        target_table (str): Таблица, на которую указывает внешний ключ.
        on_delete (str): Поведение при удалении связанной записи.
    """
    with connection.cursor() as cursor:
        # Проверяем существование таблиц
        cursor.execute("""
            SELECT EXISTS (
                SELECT FROM information_schema.tables WHERE table_name = %s AND table_schema = 'public'
            ) AND EXISTS (
                SELECT FROM information_schema.tables WHERE table_name = %s AND table_schema = 'public'
            );
        """, [table_name, target_table])

        if not cursor.fetchone()[0]:
            return

        # Найти имя constraint'а
        cursor.execute(f"""
            SELECT conname
            FROM pg_constraint
            WHERE conrelid = '{table_name}'::regclass
              AND confrelid = '{target_table}'::regclass
              AND contype = 'f'
        """)
        row = cursor.fetchone()
        if not row:
            return

        constraint_name = row[0]

        # Удалить старый constraint, Добавить новый constraint с нужным ON DELETE
        cursor.execute(f'''
            ALTER TABLE "{table_name}"
            DROP CONSTRAINT IF EXISTS "{constraint_name}",
            ADD CONSTRAINT "{constraint_name}"
            FOREIGN KEY ("{field_name}")
            REFERENCES "{target_table}"(id)
            ON DELETE {on_delete};
        ''')



from functools import (
    cached_property,
)
from io import (
    StringIO,
)
from typing import (
    Optional,
)

from django.db.models import (
    Manager,
)
from uploader_client.models import (
    Entry,
)

from edu_rdm_integration.stages.upload_data.uploader_log.enums import (
    RequestResultStatus,
)
from edu_rdm_integration.stages.upload_data.uploader_log.managers import (
    UploaderClientLogManager,
)


class UploaderClientLog(Entry):
    """Прокси модель Загрузчика данных в витрину."""

    objects = UploaderClientLogManager()
    base_objects = Manager()

    @cached_property
    def http_method_and_url(self) -> tuple[str, str]:
        """Возвращает http-метод и url из поля запроса Entry.request."""
        request = StringIO(self.request)
        request_first_line = request.readline()
        request.close()

        method, url = request_first_line.split(' ')[:2]
        if not (method and url.startswith('http')):
            method = url = ''

        return method.strip('[]'), url

    @cached_property
    def http_response_status(self) -> Optional[str]:
        """Статус-код запроса к витрине."""
        try:
            http_status = self.response.split(' ')[0].strip('[]')

            if not http_status:
                return None

            if 200 <= int(http_status) < 300:
                http_status = 'Успех'
            elif int(http_status) >= 400:
                http_status = 'Ошибка'
        except (IndexError, AttributeError):
            http_status = None

        return http_status

    @property
    def http_method(self) -> str:
        """Значение http-метода."""
        return self.http_method_and_url[0]

    @property
    def request_url(self) -> str:
        """URL запроса."""
        return self.http_method_and_url[1]

    @property
    def request_error(self) -> Optional[str]:
        """Ошибка запроса."""
        return self.error

    @property
    def result_status_display(self) -> str:
        """Результат запроса."""
        result_status = getattr(self, 'result_status', RequestResultStatus.ERROR)

        return RequestResultStatus.values.get(result_status) or RequestResultStatus.values[RequestResultStatus.ERROR]

    class Meta:
        proxy = True



# Generated by Django 3.2.24 on 2025-04-17 12:43

from django.db import (
    migrations,
    models,
)


class Migration(migrations.Migration):
    dependencies = [
        ('edu_rdm_integration', '0015_set_exporting_sub_stage_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='transferredentity',
            name='queue_level',
            field=models.PositiveIntegerField(
                choices=[(1, 'Быстрая'), (2, 'Основная'), (3, 'Долгая')], default=2, verbose_name='Уровень очереди'
            ),
        ),
    ]



from django.apps import (
    AppConfig,
)


class RDMCleanupOutdatedDataAppConfig(AppConfig):
    """Приложение для организации процесса очистки устаревших данных РВД."""

    name = 'edu_rdm_integration.pipelines.cleanup_outdated_data'
    label = 'edu_rdm_integration_cleanup_outdated_data_pipeline'
    verbose_name = 'Приложение для организации процесса очистки устаревших данных РВД.'



from abc import (
    ABCMeta,
)
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    Iterator,
)

from edu_function_tools.managers import (
    EduRunnerManager,
)
from educommon import (
    logger,
)
from m3_db_utils.models import (
    ModelEnumValue,
)

from edu_rdm_integration.core.consts import (
    LOGS_DELIMITER,
)
from edu_rdm_integration.core.storages import (
    RegionalDataMartEntityStorage,
)
from edu_rdm_integration.stages.collect_data.models import (
    RDMCollectingDataStageStatus,
)
from edu_rdm_integration.stages.export_data.functions.base.runners import (
    BaseExportDataRunner,
)
from edu_rdm_integration.stages.export_data.models import (
    RDMExportingDataStage,
    RDMExportingDataStageStatus,
)


class BaseExportDataRunnerManager(EduRunnerManager, metaclass=ABCMeta):
    """Менеджер ранеров функций выгрузки данных для интеграции с "Региональная витрина данных"."""

    forced_run = True

    def __init__(
        self,
        *args,
        is_only_main_model: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Выгрузка данных производится только для основной модели сущности
        self._is_only_main_model = is_only_main_model

        self._period_started_at, self._period_ended_at = self._prepare_period(*args, **kwargs)

        self._stage = RDMExportingDataStage.objects.create(
            manager_id=self.uuid,
            period_started_at=self._period_started_at,
            period_ended_at=self._period_ended_at,
        )

        logger.info(f'{LOGS_DELIMITER}{repr(self._stage)} created.')

    @classmethod
    def _prepare_runner_class(cls) -> type[BaseExportDataRunner]:
        """Возвращает класс ранера."""
        return BaseExportDataRunner

    def _find_exporting_data_stage(self):
        """Поиск последнего подэтапа выгрузки данных сущности РВД."""
        entity_storage = RegionalDataMartEntityStorage()
        entity_storage.prepare()

        exporting_data_stage = RDMExportingDataStage.objects.filter(
            manager_id=self.uuid,
            status_id__in=(
                RDMCollectingDataStageStatus.FAILED.key,
                RDMCollectingDataStageStatus.FINISHED.key,
            ),
        ).latest('period_ended_at')

        if exporting_data_stage:
            logger.info(f'{LOGS_DELIMITER}{repr(exporting_data_stage)} sub stages found.')
        else:
            logger.info(f'{LOGS_DELIMITER} sub stages not found.')

        return exporting_data_stage

    def _prepare_period(self, *args, **kwargs) -> tuple[datetime, datetime]:
        """Формирование периода сбора данных моделей РВД."""
        period_started_at = kwargs.get('period_started_at')
        period_ended_at = kwargs.get('period_ended_at')

        if not period_started_at:
            last_exporting_data_stage = self._find_exporting_data_stage()

            if last_exporting_data_stage:
                period_started_at = last_exporting_data_stage.period_ended_at
            else:
                period_started_at = datetime.combine(date.today(), time.min)

        if not period_ended_at:
            period_ended_at = datetime.combine(date.today(), time.max)

        return period_started_at, period_ended_at

    def _prepare_model_ids_map(self) -> dict[ModelEnumValue, Iterator[int]]:
        """Осуществляется поиск записей моделей добавленных или обновленных за указанный период времени.

        Т.к. моделей влияющих на сущность может быть множество, то в методе формируется словарь, содержащий в качестве
        ключа название модели, значение - итератор идентификаторов записей.
        """
        return {}

    def _create_runner(self, *args, **kwargs):
        """Производится расширение для осуществления поиска идентификаторов записей моделей РВД для дальнейшей выгрузки.

        model_ids_map пробрасывается в ранер для дальнейшей обработки записей для формирования чанков.
        is_force_fill_cache указывается для отказа от заполнения кешей запускаемых объектов при их создании.
        """
        model_ids_map = self._prepare_model_ids_map()

        self._stage.status_id = RDMExportingDataStageStatus.IN_PROGRESS.key
        self._stage.save()

        super()._create_runner(
            *args,
            model_ids_map=model_ids_map,
            stage=self._stage,
            is_force_fill_cache=False,
            **kwargs,
        )

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')

    def _start_runner(self, *args, **kwargs):
        """Ранер необходимо запустить с отложенным заполнением кешей.

        Это позволит произвести заполнение перед запуском объекта.
        """
        super()._start_runner(*args, is_force_fill_cache=False, **kwargs)

    def _after_start_runner(self, *args, **kwargs):
        """Точка расширения поведения менеджера ранера после запуска ранера."""
        if self._runner.result.has_not_errors:
            self._stage.status_id = RDMExportingDataStageStatus.FINISHED.key
        else:
            self._stage.status_id = RDMExportingDataStageStatus.FAILED.key

        self._stage.save()

        logger.info(f'{LOGS_DELIMITER}change status {repr(self._stage)}')



from typing import (
    Optional,
)

from m3.db import (
    BaseEnumerate,
)


class ValueCodeEnumerate(BaseEnumerate):
    """Добавляет к перечислению метод возвращающий наименование кода по константе."""

    _codes_cache = None
    _values_cache = None

    @classmethod
    def get_code_by_constant(cls, constant: int, delimiter: Optional[str] = None) -> Optional[str]:
        """Возвращает наименование кода по значению константы."""
        if cls._codes_cache is None:
            cls._codes_cache = {
                v: k.lower().replace('_', delimiter) if delimiter else k.lower()
                for k, v in cls.__dict__.items()
                if isinstance(v, int) and v in cls.values
            }

        return cls._codes_cache.get(constant)

    @classmethod
    def get_value_by_description(cls, name: str, default: Optional[int] = None) -> Optional[int]:
        """Возвращает значение по описанию этого значения."""
        if cls._values_cache is None:
            cls._values_cache = {v: k for k, v in cls.values.items() if isinstance(v, str)}

        return cls._values_cache.get(name, default)


class EntityStatusEnum(ValueCodeEnumerate):
    """Статус сущности."""

    ACTIVE = 1
    CANCELLED = 2
    ENTERED_IN_ERROR = 3

    values = {
        ACTIVE: 'Действует',
        CANCELLED: 'Отменено',
        ENTERED_IN_ERROR: 'Внесено по ошибке',
    }


class CommandType(BaseEnumerate):
    """Тип команды."""

    AUTO = 1
    MANUAL = 2

    values = {
        AUTO: 'Автоматический',
        MANUAL: 'Ручной',
    }


class AddressUse(ValueCodeEnumerate):
    """Использование адреса."""

    HOME = 1
    WORK = 2
    TEMP = 3
    OLD = 4
    BILLING = 5

    values = {
        HOME: 'Домашний',
        WORK: 'Рабочий',
        TEMP: 'Временный',
        OLD: 'Устаревший',
        BILLING: 'Для счетов',
    }


class AddressType(ValueCodeEnumerate):
    """Тип адреса."""

    POSTAL = 1
    RESIDENCE = 2
    TEMPORARY = 3
    LOCATION = 4
    JURISTICAL = 5

    values = {
        POSTAL: 'Почтовый',
        RESIDENCE: 'Место жительства',
        TEMPORARY: 'Место пребывания',
        LOCATION: 'Местонахождение',
        JURISTICAL: 'Юридический',
    }

    active_types = (POSTAL, LOCATION, JURISTICAL)


class TelecomTypeEnum(ValueCodeEnumerate):
    """Тип контактных данных."""

    PHONE = 1
    FAX = 2
    EMAIL = 3
    PAGER = 4
    URL = 5
    SMS = 6
    OTHER = 7

    values = {
        PHONE: 'Телефон',
        FAX: 'Факс',
        EMAIL: 'Электронная почта',
        PAGER: 'Пейджер',
        URL: 'Ссылка',
        SMS: 'СМС',
        OTHER: 'Иное',
    }


class TelecomUseEnum(ValueCodeEnumerate):
    """Код использования контактных данных."""

    HOME = 1
    WORK = 2
    TEMP = 3
    OLD = 4
    MOBILE = 5

    values = {
        HOME: 'Домашний',
        WORK: 'Рабочий',
        TEMP: 'Временный',
        OLD: 'Устаревший',
        MOBILE: 'Мобильный',
    }


class EntityRetiredStatusEnum(ValueCodeEnumerate):
    """Статус сущности действие которой может быть прекращено."""

    ACTIVE = 1
    RETIRED = 2
    ENTERED_IN_ERROR = 3

    values = {
        ACTIVE: 'Действует',
        RETIRED: 'Прекращена',
        ENTERED_IN_ERROR: 'Создана по ошибке',
    }



from time import sleep
from typing import (
    Any,
)

from django.core.management.base import (
    BaseCommand,
)

from django.core.cache import (
    cache,
)

from edu_rdm_integration.stages.upload_data.operations import (
    UploadData,
)
from edu_rdm_integration.stages.upload_data.queues import (
    RdmDictBasedSubStageAttachmentQueue
)


class Command(BaseCommand):
    """Команда для отправки данных в витрину параллельно-последовательно. В рамках скрипта последовательно,
    параллельно количетвом запусков команды."""

    help = 'Команда для отправки данных в витрину параллельно-последовательно'  # noqa: A003


    def handle(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> None:
        """Обработчик команды."""
        while True:
            self.stdout.write(f'Начало отправки данных в витрину')

            queue = RdmDictBasedSubStageAttachmentQueue()
            upload_data = UploadData(
                data_cache=cache,
                queue=queue,
            )

            upload_result = upload_data.upload_data()

            sleep(40)

            self.stdout.write(f'Общий объем отправленных файлов {upload_result["total_file_size"]}')
            self.stdout.write(f'Сущности, отправленные в витрину {upload_result["uploaded_entities"]}')


from edu_function_tools.results import (
    EduFunctionResult,
    EduRunnerResult,
)


class BaseCollectingCalculatedExportedDataRunnerResult(EduRunnerResult):
    """Базовый класс результатов работы ранеров функций сбора расчетных данных для РВД."""


class BaseCollectingCalculatedExportedDataFunctionResult(EduFunctionResult):
    """Базовый класс результатов работы функций сбора расчетных данных для интеграции с РВД."""
